import unittest

from library.analyses.equations.linear import linear_equation
from library.analyses.equations.quadratic import quadratic_equation
from library.analyses.equations.cubic import cubic_equation
from library.analyses.equations.hyperbolic import hyperbolic_equation
from library.analyses.equations.exponential import exponential_equation
from library.analyses.equations.logarithmic import logarithmic_equation
from library.analyses.equations.logistic import logistic_equation
from library.analyses.equations.sinusoidal import sinusoidal_equation
from library.analyses.roots.linear import linear_roots, linear_roots_first_derivative, linear_roots_second_derivative, linear_roots_initial_value, linear_roots_derivative_initial_value
from library.analyses.roots.quadratic import quadratic_roots, quadratic_roots_first_derivative, quadratic_roots_second_derivative, quadratic_roots_initial_value, quadratic_roots_derivative_initial_value
from library.analyses.roots.cubic import cubic_roots, cubic_roots_first_derivative, cubic_roots_second_derivative, cubic_roots_initial_value, cubic_roots_derivative_initial_value
from library.analyses.roots.hyperbolic import hyperbolic_roots, hyperbolic_roots_first_derivative, hyperbolic_roots_second_derivative, hyperbolic_roots_initial_value, hyperbolic_roots_derivative_initial_value
from library.analyses.roots.exponential import exponential_roots, exponential_roots_first_derivative, exponential_roots_second_derivative, exponential_roots_initial_value, exponential_roots_derivative_initial_value
from library.analyses.roots.logarithmic import logarithmic_roots, logarithmic_roots_first_derivative, logarithmic_roots_second_derivative, logarithmic_roots_initial_value, logarithmic_roots_derivative_initial_value
from library.analyses.roots.logistic import logistic_roots, logistic_roots_first_derivative, logistic_roots_second_derivative, logistic_roots_initial_value, logistic_roots_derivative_initial_value
from library.analyses.roots.sinusoidal import sinusoidal_roots, sinusoidal_roots_first_derivative, sinusoidal_roots_second_derivative, sinusoidal_roots_initial_value, sinusoidal_roots_derivative_initial_value
from library.analyses.derivatives.linear import linear_derivatives
from library.analyses.derivatives.quadratic import quadratic_derivatives
from library.analyses.derivatives.cubic import cubic_derivatives
from library.analyses.derivatives.hyperbolic import hyperbolic_derivatives
from library.analyses.derivatives.exponential import exponential_derivatives
from library.analyses.derivatives.logarithmic import logarithmic_derivatives
from library.analyses.derivatives.logistic import logistic_derivatives
from library.analyses.derivatives.sinusoidal import sinusoidal_derivatives
from library.analyses.integrals.linear import linear_integral
from library.analyses.integrals.quadratic import quadratic_integral
from library.analyses.integrals.cubic import cubic_integral
from library.analyses.integrals.hyperbolic import hyperbolic_integral
from library.analyses.integrals.exponential import exponential_integral
from library.analyses.integrals.logarithmic import logarithmic_integral
from library.analyses.integrals.logistic import logistic_integral
from library.analyses.integrals.sinusoidal import sinusoidal_integral
from library.analyses.criticals import critical_points
from library.analyses.intervals import sign_chart
from library.analyses.intercepts import intercept_points
from library.analyses.maxima import maxima_points
from library.analyses.minima import minima_points
from library.analyses.extrema import extrema_points
from library.analyses.inflections import inflection_points
from library.analyses.points import coordinate_pairs, key_coordinates, points_within_range, shifted_points_within_range, shifted_coordinates_within_range
from library.analyses.accumulation import accumulated_area
from library.analyses.mean_values import average_values

coefficients = [2, 3, 5, 7]

class TestEquations(unittest.TestCase):
    # LINEAR EVALUATIONS
    def test_linear_function_positive(self):
        linear_function_positive = linear_equation(coefficients[0], coefficients[1])(1)
        self.assertEqual(linear_function_positive, 5.0)
    
    def test_linear_function_zero(self):
        linear_function_zero = linear_equation(coefficients[0], coefficients[1])(0)
        self.assertEqual(linear_function_zero, 3.0)
    
    def test_linear_function_negative(self):
        linear_function_negative = linear_equation(coefficients[0], coefficients[1])(-1)
        self.assertEqual(linear_function_negative, 1.0)
    
    # QUADRATIC EVALUATIONS
    def test_quadratic_function_positive(self):
        quadratic_function_positive = quadratic_equation(coefficients[0], coefficients[1], coefficients[2])(1)
        self.assertEqual(quadratic_function_positive, 10.0)
    
    def test_quadratic_function_zero(self):
        quadratic_function_zero = quadratic_equation(coefficients[0], coefficients[1], coefficients[2])(0)
        self.assertEqual(quadratic_function_zero, 5.0)
    
    def test_quadratic_function_negative(self):
        quadratic_function_negative = quadratic_equation(coefficients[0], coefficients[1], coefficients[2])(-1)
        self.assertEqual(quadratic_function_negative, 4.0)
    
    # CUBIC EVALUATIONS
    def test_cubic_function_positive(self):
        cubic_function_positive = cubic_equation(coefficients[0], coefficients[1], coefficients[2], coefficients[3])(1)
        self.assertEqual(cubic_function_positive, 17.0)
    
    def test_cubic_function_zero(self):
        cubic_function_zero = cubic_equation(coefficients[0], coefficients[1], coefficients[2], coefficients[3])(0)
        self.assertEqual(cubic_function_zero, 7.0)
    
    def test_cubic_function_negative(self):
        cubic_function_negative = cubic_equation(coefficients[0], coefficients[1], coefficients[2], coefficients[3])(-1)
        self.assertEqual(cubic_function_negative, 3.0)
    
    # HYPERBOLIC EVALUATIONS
    def test_hyperbolic_function_positive(self):
        hyperbolic_function_positive = hyperbolic_equation(coefficients[0], coefficients[1])(1)
        self.assertEqual(hyperbolic_function_positive, 5.0)
    
    def test_hyperbolic_function_zero(self):
        hyperbolic_function_zero = hyperbolic_equation(coefficients[0], coefficients[1])(0)
        self.assertEqual(hyperbolic_function_zero, 20003.0)
    
    def test_hyperbolic_function_negative(self):
        hyperbolic_function_negative = hyperbolic_equation(coefficients[0], coefficients[1])(-1)
        self.assertEqual(hyperbolic_function_negative, 1.0)
    
    # EXPONENTIAL EVALUATIONS
    def test_exponential_function_positive(self):
        exponential_function_positive = exponential_equation(coefficients[0], coefficients[1])(1)
        self.assertEqual(exponential_function_positive, 6.0)
    
    def test_exponential_function_zero(self):
        exponential_function_zero = exponential_equation(coefficients[0], coefficients[1])(0)
        self.assertEqual(exponential_function_zero, 2.0)
    
    def test_exponential_function_negative(self):
        exponential_function_negative = exponential_equation(coefficients[0], coefficients[1])(-1)
        self.assertEqual(exponential_function_negative, 0.6667)
    
    # LOGARITHMIC EVALUATIONS
    def test_logarithmic_function_positive(self):
        logarithmic_function_positive = logarithmic_equation(coefficients[0], coefficients[1])(1)
        self.assertEqual(logarithmic_function_positive, 3.0)
    
    def test_logarithmic_function_zero(self):
        logarithmic_function_zero = logarithmic_equation(coefficients[0], coefficients[1])(0)
        self.assertEqual(logarithmic_function_zero, -15.4207)
    
    def test_logarithmic_function_negative(self):
        logarithmic_function_negative = logarithmic_equation(coefficients[0], coefficients[1])(-1)
        self.assertEqual(logarithmic_function_negative, 3.0)
    
    # LOGISTIC EVALUATIONS
    def test_logistic_function_positive(self):
        logistic_function_positive = logistic_equation(coefficients[0], coefficients[1], coefficients[2])(1)
        self.assertEqual(logistic_function_positive, 0.0001)
    
    def test_logistic_function_zero(self):
        logistic_function_zero = logistic_equation(coefficients[0], coefficients[1], coefficients[2])(0)
        self.assertEqual(logistic_function_zero, 0.0001)
    
    def test_logistic_function_negative(self):
        logistic_function_negative = logistic_equation(coefficients[0], coefficients[1], coefficients[2])(-1)
        self.assertEqual(logistic_function_negative, 0.0001)
    
    # SINUSOIDAL EVALUATIONS
    def test_sinusoidal_function_positive(self):
        sinusoidal_function_positive = sinusoidal_equation(coefficients[0], coefficients[1], coefficients[2], coefficients[3])(1)
        self.assertEqual(sinusoidal_function_positive, 8.0731)
    
    def test_sinusoidal_function_zero(self):
        sinusoidal_function_zero = sinusoidal_equation(coefficients[0], coefficients[1], coefficients[2], coefficients[3])(0)
        self.assertEqual(sinusoidal_function_zero, 5.6994)
    
    def test_sinusoidal_function_negative(self):
        sinusoidal_function_negative = sinusoidal_equation(coefficients[0], coefficients[1], coefficients[2], coefficients[3])(-1)
        self.assertEqual(sinusoidal_function_negative, 8.502)

class TestDerivatives(unittest.TestCase):
    # LINEAR FIRST DERIVATIVE
    def test_linear_first_derivative_constants(self):
        linear_first_derivative_constants = linear_derivatives(coefficients[0], coefficients[1])['first']['constants']
        self.assertEqual(linear_first_derivative_constants, [2.0])

    def test_linear_first_derivative_evaluation_positive(self):
        linear_first_derivative_evaluation_positive = linear_derivatives(coefficients[0], coefficients[1])['first']['evaluation'](1)
        self.assertEqual(linear_first_derivative_evaluation_positive, 2.0)
    
    def test_linear_first_derivative_evaluation_zero(self):
        linear_first_derivative_evaluation_zero = linear_derivatives(coefficients[0], coefficients[1])['first']['evaluation'](0)
        self.assertEqual(linear_first_derivative_evaluation_zero, 2.0)
    
    def test_linear_first_derivative_evaluation_negative(self):
        linear_first_derivative_evaluation_negative = linear_derivatives(coefficients[0], coefficients[1])['first']['evaluation'](-1)
        self.assertEqual(linear_first_derivative_evaluation_negative, 2.0)
    
    # QUADRATIC FIRST DERIVATIVE
    def test_quadratic_first_derivative_constants(self):
        quadratic_first_derivative_constants = quadratic_derivatives(coefficients[0], coefficients[1], coefficients[2])['first']['constants']
        self.assertEqual(quadratic_first_derivative_constants, [4.0, 3.0])
    
    def test_quadratic_first_derivative_evaluation_positive(self):
        quadratic_first_derivative_evaluation_positive = quadratic_derivatives(coefficients[0], coefficients[1], coefficients[2])['first']['evaluation'](1)
        self.assertEqual(quadratic_first_derivative_evaluation_positive, 7.0)
    
    def test_quadratic_first_derivative_evaluation_zero(self):
        quadratic_first_derivative_evaluation_zero = quadratic_derivatives(coefficients[0], coefficients[1], coefficients[2])['first']['evaluation'](0)
        self.assertEqual(quadratic_first_derivative_evaluation_zero, 3.0)
    
    def test_quadratic_first_derivative_evaluation_negative(self):
        quadratic_first_derivative_evaluation_negative = quadratic_derivatives(coefficients[0], coefficients[1], coefficients[2])['first']['evaluation'](-1)
        self.assertEqual(quadratic_first_derivative_evaluation_negative, -1.0)

    # CUBIC FIRST DERIVATIVE
    def test_cubic_first_derivative_constants(self):
        cubic_first_derivative_constants = cubic_derivatives(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['first']['constants']
        self.assertEqual(cubic_first_derivative_constants, [6.0, 6.0, 5.0])
    
    def test_cubic_first_derivative_evaluation_positive(self):
        cubic_first_derivative_evaluation_positive = cubic_derivatives(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['first']['evaluation'](1)
        self.assertEqual(cubic_first_derivative_evaluation_positive, 17.0)
    
    def test_cubic_first_derivative_evaluation_zero(self):
        cubic_first_derivative_evaluation_zero = cubic_derivatives(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['first']['evaluation'](0)
        self.assertEqual(cubic_first_derivative_evaluation_zero, 5.0)
    
    def test_cubic_first_derivative_evaluation_negative(self):
        cubic_first_derivative_evaluation_negative = cubic_derivatives(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['first']['evaluation'](-1)
        self.assertEqual(cubic_first_derivative_evaluation_negative, 5.0)

    # HYPERBOLIC FIRST DERIVATIVE
    def test_hyperbolic_first_derivative_constants(self):
        hyperbolic_first_derivative_constants = hyperbolic_derivatives(coefficients[0], coefficients[1])['first']['constants']
        self.assertEqual(hyperbolic_first_derivative_constants, [-2.0])
    
    def test_hyperbolic_first_derivative_evaluation_positive(self):
        hyperbolic_first_derivative_evaluation_positive = hyperbolic_derivatives(coefficients[0], coefficients[1])['first']['evaluation'](1)
        self.assertEqual(hyperbolic_first_derivative_evaluation_positive, -2.0)
    
    def test_hyperbolic_first_derivative_evaluation_zero(self):
        hyperbolic_first_derivative_evaluation_zero = hyperbolic_derivatives(coefficients[0], coefficients[1])['first']['evaluation'](0)
        self.assertEqual(hyperbolic_first_derivative_evaluation_zero, -200000000.0)
    
    def test_hyperbolic_first_derivative_evaluation_negative(self):
        hyperbolic_first_derivative_evaluation_negative = hyperbolic_derivatives(coefficients[0], coefficients[1])['first']['evaluation'](-1)
        self.assertEqual(hyperbolic_first_derivative_evaluation_negative, -2.0)

    # EXPONENTIAL FIRST DERIVATIVE
    def test_exponential_first_derivative_constants(self):
        exponential_first_derivative_constants = exponential_derivatives(coefficients[0], coefficients[1])['first']['constants']
        self.assertEqual(exponential_first_derivative_constants, [2.1972, 3.0])
    
    def test_exponential_first_derivative_evaluation_positive(self):
        exponential_first_derivative_evaluation_positive = exponential_derivatives(coefficients[0], coefficients[1])['first']['evaluation'](1)
        self.assertEqual(exponential_first_derivative_evaluation_positive, 6.5916)
    
    def test_exponential_first_derivative_evaluation_zero(self):
        exponential_first_derivative_evaluation_zero = exponential_derivatives(coefficients[0], coefficients[1])['first']['evaluation'](0)
        self.assertEqual(exponential_first_derivative_evaluation_zero, 2.1972)
    
    def test_exponential_first_derivative_evaluation_negative(self):
        exponential_first_derivative_evaluation_negative = exponential_derivatives(coefficients[0], coefficients[1])['first']['evaluation'](-1)
        self.assertEqual(exponential_first_derivative_evaluation_negative, 0.7324)
    
    # LOGARITHMIC FIRST DERIVATIVE
    def test_logarithmic_first_derivative_constants(self):
        logarithmic_first_derivative_constants = logarithmic_derivatives(coefficients[0], coefficients[1])['first']['constants']
        self.assertEqual(logarithmic_first_derivative_constants, [2.0])
    
    def test_logarithmic_first_derivative_evaluation_positive(self):
        logarithmic_first_derivative_evaluation_positive = logarithmic_derivatives(coefficients[0], coefficients[1])['first']['evaluation'](1)
        self.assertEqual(logarithmic_first_derivative_evaluation_positive, 2.0)
    
    def test_logarithmic_first_derivative_evaluation_zero(self):
        logarithmic_first_derivative_evaluation_zero = logarithmic_derivatives(coefficients[0], coefficients[1])['first']['evaluation'](0)
        self.assertEqual(logarithmic_first_derivative_evaluation_zero, 20000.0)
    
    def test_logarithmic_first_derivative_evaluation_negative(self):
        logarithmic_first_derivative_evaluation_negative = logarithmic_derivatives(coefficients[0], coefficients[1])['first']['evaluation'](-1)
        self.assertEqual(logarithmic_first_derivative_evaluation_negative, -2.0)
    
    # LOGISTIC FIRST DERIVATIVE
    def test_logistic_first_derivative_constants(self):
        logistic_first_derivative_constants = logistic_derivatives(coefficients[0], coefficients[1], coefficients[2])['first']['constants']
        self.assertEqual(logistic_first_derivative_constants, [6.0, 3.0, 5.0])
    
    def test_logistic_first_derivative_evaluation_positive(self):
        logistic_first_derivative_evaluation_positive = logistic_derivatives(coefficients[0], coefficients[1], coefficients[2])['first']['evaluation'](1)
        self.assertEqual(logistic_first_derivative_evaluation_positive, 0.0001)
    
    def test_logistic_first_derivative_evaluation_zero(self):
        logistic_first_derivative_evaluation_zero = logistic_derivatives(coefficients[0], coefficients[1], coefficients[2])['first']['evaluation'](0)
        self.assertEqual(logistic_first_derivative_evaluation_zero, 0.0001)
    
    def test_logistic_first_derivative_evaluation_negative(self):
        logistic_first_derivative_evaluation_negative = logistic_derivatives(coefficients[0], coefficients[1], coefficients[2])['first']['evaluation'](-1)
        self.assertEqual(logistic_first_derivative_evaluation_negative, 0.0001)

    # SINUSOIDAL FIRST DERIVATIVE
    def test_sinusoidal_first_derivative_constants(self):
        sinusoidal_first_derivative_constants = sinusoidal_derivatives(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['first']['constants']
        self.assertEqual(sinusoidal_first_derivative_constants, [6.0, 3.0, 5.0])
    
    def test_sinusoidal_first_derivative_evaluation_positive(self):
        sinusoidal_first_derivative_evaluation_positive = sinusoidal_derivatives(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['first']['evaluation'](1)
        self.assertEqual(sinusoidal_first_derivative_evaluation_positive, 5.0631)
    
    def test_sinusoidal_first_derivative_evaluation_zero(self):
        sinusoidal_first_derivative_evaluation_zero = sinusoidal_derivatives(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['first']['evaluation'](0)
        self.assertEqual(sinusoidal_first_derivative_evaluation_zero, -4.5581)
    
    def test_sinusoidal_first_derivative_evaluation_negative(self):
        sinusoidal_first_derivative_evaluation_negative = sinusoidal_derivatives(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['first']['evaluation'](-1)
        self.assertEqual(sinusoidal_first_derivative_evaluation_negative, 3.9619)

    # LINEAR SECOND DERIVATIVE
    def test_linear_second_derivative_constants(self):
        linear_second_derivative_constants = linear_derivatives(coefficients[0], coefficients[1])['second']['constants']
        self.assertEqual(linear_second_derivative_constants, [0])
    
    def test_linear_second_derivative_evaluation_positive(self):
        linear_second_derivative_evaluation_positive = linear_derivatives(coefficients[0], coefficients[1])['second']['evaluation'](1)
        self.assertEqual(linear_second_derivative_evaluation_positive, 0.0)
    
    def test_linear_second_derivative_evaluation_zero(self):
        linear_second_derivative_evaluation_zero = linear_derivatives(coefficients[0], coefficients[1])['second']['evaluation'](0)
        self.assertEqual(linear_second_derivative_evaluation_zero, 0.0)
    
    def test_linear_second_derivative_evaluation_negative(self):
        linear_second_derivative_evaluation_negative = linear_derivatives(coefficients[0], coefficients[1])['second']['evaluation'](-1)
        self.assertEqual(linear_second_derivative_evaluation_negative, 0.0)
    
    # QUADRATIC SECOND DERIVATIVE
    def test_quadratic_second_derivative_constants(self):
        quadratic_second_derivative_constants = quadratic_derivatives(coefficients[0], coefficients[1], coefficients[2])['second']['constants']
        self.assertEqual(quadratic_second_derivative_constants, [4])
    
    def test_quadratic_second_derivative_evaluation_positive(self):
        quadratic_second_derivative_evaluation_positive = quadratic_derivatives(coefficients[0], coefficients[1], coefficients[2])['second']['evaluation'](1)
        self.assertEqual(quadratic_second_derivative_evaluation_positive, 4.0)
    
    def test_quadratic_second_derivative_evaluation_zero(self):
        quadratic_second_derivative_evaluation_zero = quadratic_derivatives(coefficients[0], coefficients[1], coefficients[2])['second']['evaluation'](0)
        self.assertEqual(quadratic_second_derivative_evaluation_zero, 4.0)
    
    def test_quadratic_second_derivative_evaluation_negative(self):
        quadratic_second_derivative_evaluation_negative = quadratic_derivatives(coefficients[0], coefficients[1], coefficients[2])['second']['evaluation'](-1)
        self.assertEqual(quadratic_second_derivative_evaluation_negative, 4.0)

    # CUBIC SECOND DERIVATIVE
    def test_cubic_second_derivative_constants(self):
        cubic_second_derivative_constants = cubic_derivatives(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['second']['constants']
        self.assertEqual(cubic_second_derivative_constants, [12, 6])
    
    def test_cubic_second_derivative_evaluation_positive(self):
        cubic_second_derivative_evaluation_positive = cubic_derivatives(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['second']['evaluation'](1)
        self.assertEqual(cubic_second_derivative_evaluation_positive, 18.0)
    
    def test_cubic_second_derivative_evaluation_zero(self):
        cubic_second_derivative_evaluation_zero = cubic_derivatives(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['second']['evaluation'](0)
        self.assertEqual(cubic_second_derivative_evaluation_zero, 6.0)
    
    def test_cubic_second_derivative_evaluation_negative(self):
        cubic_second_derivative_evaluation_negative = cubic_derivatives(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['second']['evaluation'](-1)
        self.assertEqual(cubic_second_derivative_evaluation_negative, -6.0)

    # HYPERBOLIC SECOND DERIVATIVE
    def test_hyperbolic_second_derivative_constants(self):
        hyperbolic_second_derivative_constants = hyperbolic_derivatives(coefficients[0], coefficients[1])['second']['constants']
        self.assertEqual(hyperbolic_second_derivative_constants, [4])
    
    def test_hyperbolic_second_derivative_evaluation_positive(self):
        hyperbolic_second_derivative_evaluation_positive = hyperbolic_derivatives(coefficients[0], coefficients[1])['second']['evaluation'](1)
        self.assertEqual(hyperbolic_second_derivative_evaluation_positive, 4.0)
    
    def test_hyperbolic_second_derivative_evaluation_zero(self):
        hyperbolic_second_derivative_evaluation_zero = hyperbolic_derivatives(coefficients[0], coefficients[1])['second']['evaluation'](0)
        self.assertEqual(hyperbolic_second_derivative_evaluation_zero, 3999999999999.9995)
    
    def test_hyperbolic_second_derivative_evaluation_negative(self):
        hyperbolic_second_derivative_evaluation_negative = hyperbolic_derivatives(coefficients[0], coefficients[1])['second']['evaluation'](-1)
        self.assertEqual(hyperbolic_second_derivative_evaluation_negative, -4.0)

    # EXPONENTIAL SECOND DERIVATIVE
    def test_exponential_second_derivative_constants(self):
        exponential_second_derivative_constants = exponential_derivatives(coefficients[0], coefficients[1])['second']['constants']
        self.assertEqual(exponential_second_derivative_constants, [2.4139, 3])
    
    def test_exponential_second_derivative_evaluation_positive(self):
        exponential_second_derivative_evaluation_positive = exponential_derivatives(coefficients[0], coefficients[1])['second']['evaluation'](1)
        self.assertEqual(exponential_second_derivative_evaluation_positive, 7.2417)
    
    def test_exponential_second_derivative_evaluation_zero(self):
        exponential_second_derivative_evaluation_zero = exponential_derivatives(coefficients[0], coefficients[1])['second']['evaluation'](0)
        self.assertEqual(exponential_second_derivative_evaluation_zero, 2.4139)
    
    def test_exponential_second_derivative_evaluation_negative(self):
        exponential_second_derivative_evaluation_negative = exponential_derivatives(coefficients[0], coefficients[1])['second']['evaluation'](-1)
        self.assertEqual(exponential_second_derivative_evaluation_negative, 0.8046)
    
    # LOGARITHMIC SECOND DERIVATIVE
    def test_logarithmic_second_derivative_constants(self):
        logarithmic_second_derivative_constants = logarithmic_derivatives(coefficients[0], coefficients[1])['second']['constants']
        self.assertEqual(logarithmic_second_derivative_constants, [-2.0])
    
    def test_logarithmic_second_derivative_evaluation_positive(self):
        logarithmic_second_derivative_evaluation_positive = logarithmic_derivatives(coefficients[0], coefficients[1])['second']['evaluation'](1)
        self.assertEqual(logarithmic_second_derivative_evaluation_positive, -2.0)
    
    def test_logarithmic_second_derivative_evaluation_zero(self):
        logarithmic_second_derivative_evaluation_zero = logarithmic_derivatives(coefficients[0], coefficients[1])['second']['evaluation'](0)
        self.assertEqual(logarithmic_second_derivative_evaluation_zero, -200000000.0)
    
    def test_logarithmic_second_derivative_evaluation_negative(self):
        logarithmic_second_derivative_evaluation_negative = logarithmic_derivatives(coefficients[0], coefficients[1])['second']['evaluation'](-1)
        self.assertEqual(logarithmic_second_derivative_evaluation_negative, -2.0)
    
    # LOGISTIC SECOND DERIVATIVE
    def test_logistic_second_derivative_constants(self):
        logistic_second_derivative_constants = logistic_derivatives(coefficients[0], coefficients[1], coefficients[2])['second']['constants']
        self.assertEqual(logistic_second_derivative_constants, [18.0, 3.0, 5.0])
    
    def test_logistic_second_derivative_evaluation_positive(self):
        logistic_second_derivative_evaluation_positive = logistic_derivatives(coefficients[0], coefficients[1], coefficients[2])['second']['evaluation'](1)
        self.assertEqual(logistic_second_derivative_evaluation_positive, 0.0001)
    
    def test_logistic_second_derivative_evaluation_zero(self):
        logistic_second_derivative_evaluation_zero = logistic_derivatives(coefficients[0], coefficients[1], coefficients[2])['second']['evaluation'](0)
        self.assertEqual(logistic_second_derivative_evaluation_zero, 0.0001)
    
    def test_logistic_second_derivative_evaluation_negative(self):
        logistic_second_derivative_evaluation_negative = logistic_derivatives(coefficients[0], coefficients[1], coefficients[2])['second']['evaluation'](-1)
        self.assertEqual(logistic_second_derivative_evaluation_negative, 0.0001)

    # SINUSOIDAL SECOND DERIVATIVE
    def test_sinusoidal_second_derivative_constants(self):
        sinusoidal_second_derivative_constants = sinusoidal_derivatives(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['second']['constants']
        self.assertEqual(sinusoidal_second_derivative_constants, [-18.0, 3.0, 5.0])

    def test_sinusoidal_second_derivative_evaluation_positive(self):
        sinusoidal_second_derivative_evaluation_positive = sinusoidal_derivatives(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['second']['evaluation'](1)
        self.assertEqual(sinusoidal_second_derivative_evaluation_positive, -9.6583)
    
    def test_sinusoidal_second_derivative_evaluation_zero(self):
        sinusoidal_second_derivative_evaluation_zero = sinusoidal_derivatives(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['second']['evaluation'](0)
        self.assertEqual(sinusoidal_second_derivative_evaluation_zero, 11.7052)
    
    def test_sinusoidal_second_derivative_evaluation_negative(self):
        sinusoidal_second_derivative_evaluation_negative = sinusoidal_derivatives(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['second']['evaluation'](-1)
        self.assertEqual(sinusoidal_second_derivative_evaluation_negative, -13.5178)

class TestIntegrals(unittest.TestCase):
    # LINEAR INTEGRAL
    def test_linear_integral_constants(self):
        linear_integral_constants = linear_integral(coefficients[0], coefficients[1])['constants']
        self.assertEqual(linear_integral_constants, [1.0, 3.0])
    
    def test_linear_integral_constants_ones(self):
        linear_integral_constants_ones = linear_integral(1, 1)['constants']
        self.assertEqual(linear_integral_constants_ones, [0.5, 1.0])
    
    def test_linear_integral_evaluation_positive(self):
        linear_integral_evaluation_positive = linear_integral(coefficients[0], coefficients[1])['evaluation'](1)
        self.assertEqual(linear_integral_evaluation_positive, 4.0)
    
    def test_linear_integral_evaluation_zero(self):
        linear_integral_evaluation_zero = linear_integral(coefficients[0], coefficients[1])['evaluation'](0)
        self.assertEqual(linear_integral_evaluation_zero, 0.0)
    
    def test_linear_integral_evaluation_negative(self):
        linear_integral_evaluation_negative = linear_integral(coefficients[0], coefficients[1])['evaluation'](-1)
        self.assertEqual(linear_integral_evaluation_negative, -2.0)
    
    # QUADRATIC INTEGRAL
    def test_quadratic_integral_constants(self):
        quadratic_integral_constants = quadratic_integral(coefficients[0], coefficients[1], coefficients[2])['constants']
        self.assertEqual(quadratic_integral_constants, [0.6667, 1.5, 5.0])
    
    def test_quadratic_integral_constants_ones(self):
        quadratic_integral_constants_ones = quadratic_integral(1, 1, 1)['constants']
        self.assertEqual(quadratic_integral_constants_ones, [0.3333, 0.5, 1.0])
    
    def test_quadratic_integral_evaluation_positive(self):
        quadratic_integral_evaluation_positive = quadratic_integral(coefficients[0], coefficients[1], coefficients[2])['evaluation'](1)
        self.assertEqual(quadratic_integral_evaluation_positive, 7.1667)
    
    def test_quadratic_integral_evaluation_zero(self):
        quadratic_integral_evaluation_zero = quadratic_integral(coefficients[0], coefficients[1], coefficients[2])['evaluation'](0)
        self.assertEqual(quadratic_integral_evaluation_zero, 0.0)
    
    def test_quadratic_integral_evaluation_negative(self):
        quadratic_integral_evaluation_negative = quadratic_integral(coefficients[0], coefficients[1], coefficients[2])['evaluation'](-1)
        self.assertEqual(quadratic_integral_evaluation_negative, -4.1667)
    
    # CUBIC INTEGRAL
    def test_cubic_integral_constants(self):
        cubic_integral_constants = cubic_integral(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['constants']
        self.assertEqual(cubic_integral_constants, [0.5, 1.0, 2.5, 7.0])
    
    def test_cubic_integral_constants_ones(self):
        cubic_integral_constants_ones = cubic_integral(1, 1, 1, 1)['constants']
        self.assertEqual(cubic_integral_constants_ones, [0.25, 0.3333, 0.5, 1.0])
    
    def test_cubic_integral_evaluation_positive(self):
        cubic_integral_evaluation_positive = cubic_integral(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['evaluation'](1)
        self.assertEqual(cubic_integral_evaluation_positive, 11.0)
    
    def test_cubic_integral_evaluation_zero(self):
        cubic_integral_evaluation_zero = cubic_integral(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['evaluation'](0)
        self.assertEqual(cubic_integral_evaluation_zero, 0.0)
    
    def test_cubic_integral_evaluation_negative(self):
        cubic_integral_evaluation_negative = cubic_integral(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['evaluation'](-1)
        self.assertEqual(cubic_integral_evaluation_negative, -5.0)
    
    # HYPERBOLIC INTEGRAL
    def test_hyperbolic_integral_constants(self):
        hyperbolic_integral_constants = hyperbolic_integral(coefficients[0], coefficients[1])['constants']
        self.assertEqual(hyperbolic_integral_constants, [2.0, 3.0])
    
    def test_hyperbolic_integral_constants_ones(self):
        hyperbolic_integral_constants_ones = hyperbolic_integral(1, 1)['constants']
        self.assertEqual(hyperbolic_integral_constants_ones, [1.0, 1.0])
    
    def test_hyperbolic_integral_evaluation_positive(self):
        hyperbolic_integral_evaluation_positive = hyperbolic_integral(coefficients[0], coefficients[1])['evaluation'](1)
        self.assertEqual(hyperbolic_integral_evaluation_positive, 3.0)
    
    def test_hyperbolic_integral_evaluation_zero(self):
        hyperbolic_integral_evaluation_zero = hyperbolic_integral(coefficients[0], coefficients[1])['evaluation'](0)
        self.assertEqual(hyperbolic_integral_evaluation_zero, -18.4204)
    
    def test_hyperbolic_integral_evaluation_negative(self):
        hyperbolic_integral_evaluation_negative = hyperbolic_integral(coefficients[0], coefficients[1])['evaluation'](-1)
        self.assertEqual(hyperbolic_integral_evaluation_negative, -3.0)
    
    # EXPONENTIAL INTEGRAL
    def test_exponential_integral_constants(self):
        exponential_integral_constants = exponential_integral(coefficients[0], coefficients[1])['constants']
        self.assertEqual(exponential_integral_constants, [1.8205, 3.0])
    
    def test_exponential_integral_constants_ones(self):
        exponential_integral_constants_ones = exponential_integral(1, 1)['constants']
        self.assertEqual(exponential_integral_constants_ones, [10000.5, 1.0001])
    
    def test_exponential_integral_evaluation_positive(self):
        exponential_integral_evaluation_positive = exponential_integral(coefficients[0], coefficients[1])['evaluation'](1)
        self.assertEqual(exponential_integral_evaluation_positive, 5.4615)
    
    def test_exponential_integral_evaluation_zero(self):
        exponential_integral_evaluation_zero = exponential_integral(coefficients[0], coefficients[1])['evaluation'](0)
        self.assertEqual(exponential_integral_evaluation_zero, 1.8205)
    
    def test_exponential_integral_evaluation_negative(self):
        exponential_integral_evaluation_negative = exponential_integral(coefficients[0], coefficients[1])['evaluation'](-1)
        self.assertEqual(exponential_integral_evaluation_negative, 0.6068)
    
    # LOGARITHMIC INTEGRAL
    def test_logarithmic_integral_constants(self):
        logarithmic_integral_constants = logarithmic_integral(coefficients[0], coefficients[1])['constants']
        self.assertEqual(logarithmic_integral_constants, [2.0, 3.0])
    
    def test_logarithmic_integral_constants_ones(self):
        logarithmic_integral_constants_ones = logarithmic_integral(1, 1)['constants']
        self.assertEqual(logarithmic_integral_constants_ones, [1.0, 1.0])
    
    def test_logarithmic_integral_evaluation_positive(self):
        logarithmic_integral_evaluation_positive = logarithmic_integral(coefficients[0], coefficients[1])['evaluation'](1)
        self.assertEqual(logarithmic_integral_evaluation_positive, 1.0)
    
    def test_logarithmic_integral_evaluation_zero(self):
        logarithmic_integral_evaluation_zero = logarithmic_integral(coefficients[0], coefficients[1])['evaluation'](0)
        self.assertEqual(logarithmic_integral_evaluation_zero, -0.0017)
    
    def test_logarithmic_integral_evaluation_negative(self):
        logarithmic_integral_evaluation_negative = logarithmic_integral(coefficients[0], coefficients[1])['evaluation'](-1)
        self.assertEqual(logarithmic_integral_evaluation_negative, -1.0)
    
    # LOGISTIC INTEGRAL
    def test_logistic_integral_constants(self):
        logistic_integral_constants = logistic_integral(coefficients[0], coefficients[1], coefficients[2])['constants']
        self.assertEqual(logistic_integral_constants, [0.6667, 3.0, 5.0])
    
    def test_logistic_integral_constants_ones(self):
        logistic_integral_constants_ones = logistic_integral(1, 1, 1)['constants']
        self.assertEqual(logistic_integral_constants_ones, [1.0, 1.0, 1.0])
    
    def test_logistic_integral_evaluation_positive(self):
        logistic_integral_evaluation_positive = logistic_integral(coefficients[0], coefficients[1], coefficients[2])['evaluation'](1)
        self.assertEqual(logistic_integral_evaluation_positive, 0.0001)
    
    def test_logistic_integral_evaluation_zero(self):
        logistic_integral_evaluation_zero = logistic_integral(coefficients[0], coefficients[1], coefficients[2])['evaluation'](0)
        self.assertEqual(logistic_integral_evaluation_zero, 0.0001)
    
    def test_logistic_integral_evaluation_negative(self):
        logistic_integral_evaluation_negative = logistic_integral(coefficients[0], coefficients[1], coefficients[2])['evaluation'](-1)
        self.assertEqual(logistic_integral_evaluation_negative, 0.0001)
    
    # SINUSOIDAL INTEGRAL
    def test_sinusoidal_integral_constants(self):
        sinusoidal_integral_constants = sinusoidal_integral(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['constants']
        self.assertEqual(sinusoidal_integral_constants, [-0.6667, 3.0, 5.0, 7.0])
    
    def test_sinusoidal_integral_constants_ones(self):
        sinusoidal_integral_constants_ones = sinusoidal_integral(1, 1, 1, 1)['constants']
        self.assertEqual(sinusoidal_integral_constants_ones, [-1.0, 1.0, 1.0, 1.0])
    
    def test_sinusoidal_integral_evaluation_positive(self):
        sinusoidal_integral_evaluation_positive = sinusoidal_integral(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['evaluation'](1)
        self.assertEqual(sinusoidal_integral_evaluation_positive, 6.4374)
    
    def test_sinusoidal_integral_evaluation_zero(self):
        sinusoidal_integral_evaluation_zero = sinusoidal_integral(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['evaluation'](0)
        self.assertEqual(sinusoidal_integral_evaluation_zero, 0.5065)
    
    def test_sinusoidal_integral_evaluation_negative(self):
        sinusoidal_integral_evaluation_negative = sinusoidal_integral(coefficients[0], coefficients[1], coefficients[2], coefficients[3])['evaluation'](-1)
        self.assertEqual(sinusoidal_integral_evaluation_negative, -7.4402)

class TestCriticalPoints(unittest.TestCase):
    # FIRST DERIVATIVE CRITICAL POINTS
    def test_first_linear_critical_points(self):
        first_linear_critical_points = critical_points('linear', coefficients[:2], 1)
        self.assertEqual(first_linear_critical_points, [None])
    
    def test_first_quadratic_critical_points(self):
        first_quadratic_critical_points = critical_points('quadratic', coefficients[:3], 1)
        self.assertEqual(first_quadratic_critical_points, [-0.75])
    
    def test_first_cubic_critical_points_none(self):
        first_cubic_critical_points_none = critical_points('cubic', coefficients[:4], 1)
        self.assertEqual(first_cubic_critical_points_none, [None])
    
    def test_first_cubic_critical_points_one(self):
        first_cubic_critical_points_one = critical_points('cubic', [1, 3, 3, 7], 1)
        self.assertEqual(first_cubic_critical_points_one, [-1.0])
    
    def test_first_cubic_critical_points_two(self):
        first_cubic_critical_points_two = critical_points('cubic', [2, -15, 36, 7], 1)
        self.assertEqual(first_cubic_critical_points_two, [2.0, 3.0])
    
    def test_first_hyperbolic_critical_points(self):
        first_hyperbolic_critical_points = critical_points('hyperbolic', coefficients[:2], 1)
        self.assertEqual(first_hyperbolic_critical_points, [0])
    
    def test_first_exponential_critical_points(self):
        first_exponential_critical_points = critical_points('exponential', coefficients[:2], 1)
        self.assertEqual(first_exponential_critical_points, [None])
    
    def test_first_logarithmic_critical_points(self):
        first_logarithmic_critical_points = critical_points('logarithmic', coefficients[:2], 1)
        self.assertEqual(first_logarithmic_critical_points, [None])
    
    def test_first_logistic_critical_points(self):
        first_logistic_critical_points = critical_points('logistic', coefficients[:3], 1)
        self.assertEqual(first_logistic_critical_points, [None])
    
    def test_first_sinusoidal_critical_points(self):
        first_sinusoidal_critical_points = critical_points('sinusoidal', coefficients[:4], 1)
        self.assertEqual(first_sinusoidal_critical_points, [5.5236, 6.5708, 7.618, 8.6652, 9.7124, '5.5236 + 1.0472k'])
    
    # SECOND DERIVATIVE CRITICAL POINTS
    def test_second_linear_critical_points(self):
        second_linear_critical_points = critical_points('linear', coefficients[:2], 2)
        self.assertEqual(second_linear_critical_points, [None])
    
    def test_second_quadratic_critical_points(self):
        second_quadratic_critical_points = critical_points('quadratic', coefficients[:3], 2)
        self.assertEqual(second_quadratic_critical_points, [None])
    
    def test_second_cubic_critical_points(self):
        second_cubic_critical_points = critical_points('cubic', coefficients[:4], 2)
        self.assertEqual(second_cubic_critical_points, [-0.5])
    
    def test_second_hyperbolic_critical_points(self):
        second_hyperbolic_critical_points = critical_points('hyperbolic', coefficients[:2], 2)
        self.assertEqual(second_hyperbolic_critical_points, [0])
    
    def test_second_exponential_critical_points(self):
        second_exponential_critical_points = critical_points('exponential', coefficients[:2], 2)
        self.assertEqual(second_exponential_critical_points, [None])
    
    def test_second_logarithmic_critical_points(self):
        second_logarithmic_critical_points = critical_points('logarithmic', coefficients[:2], 2)
        self.assertEqual(second_logarithmic_critical_points, [None])
    
    def test_second_logistic_critical_points(self):
        second_logistic_critical_points = critical_points('logistic', coefficients[:3], 2)
        self.assertEqual(second_logistic_critical_points, [5])
    
    def test_second_sinusoidal_critical_points(self):
        second_sinusoidal_critical_points = critical_points('sinusoidal', coefficients[:4], 2)
        self.assertEqual(second_sinusoidal_critical_points, [5.0, 6.0472, 7.0944, 8.1416, 9.1888, '5.0 + 1.0472k'])

class TestIntervals(unittest.TestCase):
    maxDiff = None

    # FIRST DERIVATIVE SIGN CHARTS
    def test_first_linear_intervals(self):
        first_linear_intervals = sign_chart('linear', coefficients[:2], 1)
        self.assertEqual(first_linear_intervals, ['positive'])
    
    def test_first_quadratic_intervals(self):
        first_quadratic_intervals = sign_chart('quadratic', coefficients[:3], 1)
        self.assertEqual(first_quadratic_intervals, ['negative', -0.75, 'positive'])
    
    def test_first_cubic_intervals_none(self):
        first_cubic_intervals_none = sign_chart('cubic', coefficients[:4], 1)
        self.assertEqual(first_cubic_intervals_none, ['positive'])
    
    def test_first_cubic_intervals_one(self):
        first_cubic_intervals_one = sign_chart('cubic', [1, 3, 3, 7], 1)
        self.assertEqual(first_cubic_intervals_one, ['positive', -1.0, 'positive'])
    
    def test_first_cubic_intervals_two(self):
        first_cubic_intervals_two = sign_chart('cubic', [2, -15, 36, 7], 1)
        self.assertEqual(first_cubic_intervals_two, ['positive', 2.0, 'negative', 3.0, 'positive'])

    def test_first_hyperbolic_intervals(self):
        first_hyperbolic_intervals = sign_chart('hyperbolic', coefficients[:2], 1)
        self.assertEqual(first_hyperbolic_intervals, ['negative', 0, 'negative'])
    
    def test_first_exponential_intervals(self):
        first_exponential_intervals = sign_chart('exponential', coefficients[:2], 1)
        self.assertEqual(first_exponential_intervals, ['positive'])
    
    def test_first_logarithmic_intervals(self):
        first_logarithmic_intervals = sign_chart('logarithmic', coefficients[:2], 1)
        self.assertEqual(first_logarithmic_intervals, ['positive'])
    
    def test_first_logistic_intervals(self):
        first_logistic_intervals = sign_chart('logistic', coefficients[:3], 1)
        self.assertEqual(first_logistic_intervals, ['positive'])
    
    def test_first_sinusoidal_intervals(self):
        first_sinusoidal_intervals = sign_chart('sinusoidal', coefficients[:4], 1)
        self.assertEqual(first_sinusoidal_intervals, ['positive', 5.5236, 'negative', 6.5708, 'positive', 7.618, 'negative', 8.6652, 'positive', 9.7124, 'negative', '5.5236 + 1.0472k'])
    
    # SECOND DERIVATIVE SIGN CHARTS
    def test_second_linear_intervals(self):
        second_linear_intervals = sign_chart('linear', coefficients[:2], 2)
        self.assertEqual(second_linear_intervals, ['constant'])
    
    def test_second_quadratic_intervals(self):
        second_quadratic_intervals = sign_chart('quadratic', coefficients[:3], 2)
        self.assertEqual(second_quadratic_intervals, ['positive'])
    
    def test_second_cubic_intervals(self):
        second_cubic_intervals = sign_chart('cubic', coefficients[:4], 2)
        self.assertEqual(second_cubic_intervals, ['negative', -0.5, 'positive'])
    
    def test_second_hyperbolic_intervals(self):
        second_hyperbolic_intervals = sign_chart('hyperbolic', coefficients[:2], 2)
        self.assertEqual(second_hyperbolic_intervals, ['negative', 0, 'positive'])
    
    def test_second_exponential_intervals(self):
        second_exponential_intervals = sign_chart('exponential', coefficients[:2], 2)
        self.assertEqual(second_exponential_intervals, ['positive'])
    
    def test_second_logarithmic_intervals(self):
        second_logarithmic_intervals = sign_chart('logarithmic', coefficients[:2], 2)
        self.assertEqual(second_logarithmic_intervals, ['negative'])
    
    def test_second_logistic_intervals(self):
        second_logistic_intervals = sign_chart('logistic', coefficients[:3], 2)
        self.assertEqual(second_logistic_intervals, ['positive', 5, 'negative'])
    
    def test_second_sinusoidal_intervals(self):
        second_sinusoidal_intervals = sign_chart('sinusoidal', coefficients[:4], 2)
        self.assertEqual(second_sinusoidal_intervals, ['positive', 5.0, 'negative', 6.0472, 'positive', 7.0944, 'negative', 8.1416, 'positive', 9.1888, 'negative', '5.0 + 1.0472k'])

class TestRoots(unittest.TestCase):
    maxDiff = None

    # LINEAR ROOTS
    def test_linear_zeroes(self):
        linear_zeroes = linear_roots(coefficients[0], coefficients[1])
        self.assertEqual(linear_zeroes, [-1.5])
    
    def test_linear_zeroes_first_derivative(self):
        linear_zeroes_first_derivative = linear_roots_first_derivative(coefficients[0], coefficients[1])
        self.assertEqual(linear_zeroes_first_derivative, [None])
    
    def test_linear_zeroes_second_derivative(self):
        linear_zeroes_second_derivative = linear_roots_second_derivative(coefficients[0], coefficients[1])
        self.assertEqual(linear_zeroes_second_derivative, [None])
    
    def test_linear_zeroes_initial_value(self):
        linear_zeroes_initial_value = linear_roots_initial_value(coefficients[0], coefficients[1], 10)
        self.assertEqual(linear_zeroes_initial_value, [3.5])
    
    def test_linear_zeroes_derivative_initial_value_all(self):
        linear_zeroes_derivative_initial_value_all = linear_roots_derivative_initial_value(coefficients[0], coefficients[1], 2)
        self.assertEqual(linear_zeroes_derivative_initial_value_all, ['All'])
    
    def test_linear_zeroes_derivative_initial_value_none(self):
        linear_zeroes_derivative_initial_value_none = linear_roots_derivative_initial_value(coefficients[0], coefficients[1], 10)
        self.assertEqual(linear_zeroes_derivative_initial_value_none, [None])
    
    # QUADRATIC ROOTS
    def test_quadratic_zeroes_none(self):
        quadratic_zeroes_none = quadratic_roots(coefficients[0], coefficients[1], coefficients[2])
        self.assertEqual(quadratic_zeroes_none, [None])
    
    def test_quadratic_zeroes_one(self):
        quadratic_zeroes_one = quadratic_roots(1, -2, 1)
        self.assertEqual(quadratic_zeroes_one, [1.0])
    
    def test_quadratic_zeroes_two(self):
        quadratic_zeroes_two = quadratic_roots(1, -5, 6)
        self.assertEqual(quadratic_zeroes_two, [2.0, 3.0])

    def test_quadratic_zeroes_first_derivative(self):
        quadratic_zeroes_first_derivative = quadratic_roots_first_derivative(coefficients[0], coefficients[1], coefficients[2])
        self.assertEqual(quadratic_zeroes_first_derivative, [-0.75])
    
    def test_quadratic_zeroes_second_derivative(self):
        quadratic_zeroes_second_derivative = quadratic_roots_second_derivative(coefficients[0], coefficients[1], coefficients[2])
        self.assertEqual(quadratic_zeroes_second_derivative, [None])
    
    def test_quadratic_zeroes_initial_value_none(self):
        quadratic_zeroes_initial_value_none = quadratic_roots_initial_value(coefficients[0], coefficients[1], coefficients[2], -10)
        self.assertEqual(quadratic_zeroes_initial_value_none, [None])
    
    def test_quadratic_zeroes_initial_value_one(self):
        quadratic_zeroes_initial_value_one = quadratic_roots_initial_value(1, 2, 2, 1)
        self.assertEqual(quadratic_zeroes_initial_value_one, [-1.0])
    
    def test_quadratic_zeroes_initial_value_two(self):
        quadratic_zeroes_initial_value_two = quadratic_roots_initial_value(coefficients[0], coefficients[1], coefficients[2], 10)
        self.assertEqual(quadratic_zeroes_initial_value_two, [-2.5, 1.0])
    
    def test_quadratic_zeroes_derivative_initial_value(self):
        quadratic_zeroes_derivative_initial_value = quadratic_roots_derivative_initial_value(coefficients[0], coefficients[1], coefficients[2], 10)
        self.assertEqual(quadratic_zeroes_derivative_initial_value, [1.75])

    # CUBIC ROOTS
    def test_cubic_zeroes_one(self):
        cubic_zeroes_one = cubic_roots(coefficients[0], coefficients[1], coefficients[2], coefficients[3])
        self.assertEqual(cubic_zeroes_one, [-1.4455])
    
    def test_cubic_zeroes_two(self):
        cubic_zeroes_two = cubic_roots(1, -4, 5, -2)
        self.assertEqual(cubic_zeroes_two, [1.0, 2.0])
    
    def test_cubic_zeroes_three(self):
        cubic_zeroes_three = cubic_roots(1, -10, 31, -30)
        self.assertEqual(cubic_zeroes_three, [2.0, 3.0, 5.0])

    def test_cubic_zeroes_first_derivative_none(self):
        cubic_zeroes_first_derivative_none = cubic_roots_first_derivative(coefficients[0], coefficients[1], coefficients[2], coefficients[3])
        self.assertEqual(cubic_zeroes_first_derivative_none, [None])

    def test_cubic_zeroes_first_derivative_one(self):
        cubic_zeroes_first_derivative_one = cubic_roots_first_derivative(3, 3, 1, 7)
        self.assertEqual(cubic_zeroes_first_derivative_one, [-0.3333])
    
    def test_cubic_zeroes_first_derivative_two(self):
        cubic_zeroes_first_derivative_two = cubic_roots_first_derivative(2, 7, 3, 5)
        self.assertEqual(cubic_zeroes_first_derivative_two, [-2.0946, -0.2387])
    
    def test_cubic_zeroes_second_derivative(self):
        cubic_zeroes_second_derivative = cubic_roots_second_derivative(coefficients[0], coefficients[1], coefficients[2], coefficients[3])
        self.assertEqual(cubic_zeroes_second_derivative, [-0.5])
    
    def test_cubic_zeroes_initial_value_one(self):
        cubic_zeroes_initial_value_one = cubic_roots_initial_value(coefficients[0], coefficients[1], coefficients[2], coefficients[3], 10)
        self.assertEqual(cubic_zeroes_initial_value_one, [0.4455])
    
    def test_cubic_zeroes_initial_value_two(self):
        cubic_zeroes_initial_value_two = cubic_roots_initial_value(1, -4, 5, 7, 9)
        self.assertEqual(cubic_zeroes_initial_value_two, [1.0, 2.0])
    
    def test_cubic_zeroes_initial_value_three(self):
        cubic_zeroes_initial_value_three = cubic_roots_initial_value(1, -10, 31, -20, 10)
        self.assertEqual(cubic_zeroes_initial_value_three, [2.0, 3.0, 5.0])

    def test_cubic_zeroes_derivative_initial_value_none(self):
        cubic_zeroes_derivative_initial_value_none = cubic_roots_derivative_initial_value(7, 2, 5, 3, -10)
        self.assertEqual(cubic_zeroes_derivative_initial_value_none, [None])

    def test_cubic_zeroes_derivative_initial_value_one(self):
        cubic_zeroes_derivative_initial_value_one = cubic_roots_derivative_initial_value(3, 3, 11, 1, 10)
        self.assertEqual(cubic_zeroes_derivative_initial_value_one, [-0.3333])
    
    def test_cubic_zeroes_derivative_initial_value_two(self):
        cubic_zeroes_derivative_initial_value_two = cubic_roots_derivative_initial_value(coefficients[0], coefficients[1], coefficients[2], coefficients[3], 10)
        self.assertEqual(cubic_zeroes_derivative_initial_value_two, [-1.5408, 0.5408])

    # HYPERBOLIC ROOTS
    def test_hyperbolic_zeroes(self):
        hyperbolic_zeroes = hyperbolic_roots(coefficients[0], coefficients[1])
        self.assertEqual(hyperbolic_zeroes, [-0.6667])
    
    def test_hyperbolic_zeroes_first_derivative(self):
        hyperbolic_zeroes_first_derivative = hyperbolic_roots_first_derivative(coefficients[0], coefficients[1])
        self.assertEqual(hyperbolic_zeroes_first_derivative, [0.0])
    
    def test_hyperbolic_zeroes_second_derivative(self):
        hyperbolic_zeroes_second_derivative = hyperbolic_roots_second_derivative(coefficients[0], coefficients[1])
        self.assertEqual(hyperbolic_zeroes_second_derivative, [0.0])
    
    def test_hyperbolic_zeroes_initial_value(self):
        hyperbolic_zeroes_initial_value = hyperbolic_roots_initial_value(coefficients[0], coefficients[1], 10)
        self.assertEqual(hyperbolic_zeroes_initial_value, [0.2857])
    
    def test_hyperbolic_zeroes_derivative_initial_value_none(self):
        hyperbolic_zeroes_derivative_initial_value_none = hyperbolic_roots_derivative_initial_value(coefficients[0], coefficients[1], 10)
        self.assertEqual(hyperbolic_zeroes_derivative_initial_value_none, [None])
    
    def test_hyperbolic_zeroes_derivative_initial_value_one(self):
        hyperbolic_zeroes_derivative_initial_value_one = hyperbolic_roots_derivative_initial_value(coefficients[0], coefficients[1], -10)
        self.assertEqual(hyperbolic_zeroes_derivative_initial_value_one, [0.4472])

    # EXPONENTIAL ROOTS
    def test_exponential_zeroes(self):
        exponential_zeroes = exponential_roots(coefficients[0], coefficients[1])
        self.assertEqual(exponential_zeroes, [None])
    
    def test_exponential_zeroes_first_derivative(self):
        exponential_zeroes_first_derivative = exponential_roots_first_derivative(coefficients[0], coefficients[1])
        self.assertEqual(exponential_zeroes_first_derivative, [None])
    
    def test_exponential_zeroes_second_derivative(self):
        exponential_zeroes_second_derivative = exponential_roots_second_derivative(coefficients[0], coefficients[1])
        self.assertEqual(exponential_zeroes_second_derivative, [None])
    
    def test_exponential_zeroes_initial_value(self):
        exponential_zeroes_initial_value = exponential_roots_initial_value(coefficients[0], coefficients[1], 10)
        self.assertEqual(exponential_zeroes_initial_value, [1.465])
    
    def test_exponential_zeroes_initial_value_log1(self):
        exponential_zeroes_initial_value_log1 = exponential_roots_initial_value(2, 1, 10)
        self.assertEqual(exponential_zeroes_initial_value_log1, [16094.3791])
    
    def test_exponential_zeroes_derivative_initial_value(self):
        exponential_zeroes_derivative_initial_value = exponential_roots_derivative_initial_value(coefficients[0], coefficients[1], 10)
        self.assertEqual(exponential_zeroes_derivative_initial_value, [1.3794])
    
    def test_exponential_zeroes_derivative_initial_value_log1(self):
        exponential_zeroes_derivative_initial_value_log1 = exponential_roots_derivative_initial_value(2, 1, -10)
        self.assertEqual(exponential_zeroes_derivative_initial_value_log1, [108197.7828])

    # LOGARITHMIC ROOTS
    def test_logarithmic_zeroes(self):
        logarithmic_zeroes = logarithmic_roots(coefficients[0], coefficients[1])
        self.assertEqual(logarithmic_zeroes, [0.2231])
    
    def test_logarithmic_zeroes_first_derivative(self):
        logarithmic_zeroes_first_derivative = logarithmic_roots_first_derivative(coefficients[0], coefficients[1])
        self.assertEqual(logarithmic_zeroes_first_derivative, [None])
    
    def test_logarithmic_zeroes_second_derivative(self):
        logarithmic_zeroes_second_derivative = logarithmic_roots_second_derivative(coefficients[0], coefficients[1])
        self.assertEqual(logarithmic_zeroes_second_derivative, [None])
    
    def test_logarithmic_zeroes_initial_value(self):
        logarithmic_zeroes_initial_value = logarithmic_roots_initial_value(coefficients[0], coefficients[1], 10)
        self.assertEqual(logarithmic_zeroes_initial_value, [33.1155])
    
    def test_logarithmic_zeroes_derivative_initial_value(self):
        logarithmic_zeroes_derivative_initial_value = logarithmic_roots_derivative_initial_value(coefficients[0], coefficients[1], 10)
        self.assertEqual(logarithmic_zeroes_derivative_initial_value, [0.2])
    
    # LOGISTIC ROOTS
    def test_logistic_zeroes(self):
        logistic_zeroes = logistic_roots(coefficients[0], coefficients[1], coefficients[2])
        self.assertEqual(logistic_zeroes, [None])
    
    def test_logistic_zeroes_first_derivative(self):
        logistic_zeroes_first_derivative = logistic_roots_first_derivative(coefficients[0], coefficients[1], coefficients[2])
        self.assertEqual(logistic_zeroes_first_derivative, [None])
    
    def test_logistic_zeroes_second_derivative(self):
        logistic_zeroes_second_derivative = logistic_roots_second_derivative(coefficients[0], coefficients[1], coefficients[2])
        self.assertEqual(logistic_zeroes_second_derivative, [5.0])
    
    def test_logistic_zeroes_initial_value(self):
        logistic_zeroes_initial_value = logistic_roots_initial_value(coefficients[0], coefficients[1], coefficients[2], 10)
        self.assertEqual(logistic_zeroes_initial_value, [5.0744])
    
    def test_logistic_zeroes_initial_value_log0(self):
        logistic_zeroes_initial_value_log0 = logistic_roots_initial_value(1, 2, 3, 1)
        self.assertEqual(logistic_zeroes_initial_value_log0, [7.6052])
    
    def test_logistic_zeroes_derivative_initial_value_none(self):
        logistic_zeroes_derivative_initial_value_none = logistic_roots_derivative_initial_value(coefficients[0], coefficients[1], coefficients[2], 10)
        self.assertEqual(logistic_zeroes_derivative_initial_value_none, [None])
    
    def test_logistic_zeroes_derivative_initial_value_one(self):
        logistic_zeroes_derivative_initial_value_one = logistic_roots_derivative_initial_value(1, 4, 2, 1)
        self.assertEqual(logistic_zeroes_derivative_initial_value_one, [2.0])
    
    def test_logistic_zeroes_derivative_initial_value_two(self):
        logistic_zeroes_derivative_initial_value_two = logistic_roots_derivative_initial_value(7, 5, 3, 2)
        self.assertEqual(logistic_zeroes_derivative_initial_value_two, [2.4527, 3.5473])

    # SINUSOIDAL ROOTS
    def test_sinusoidal_zeroes_none(self):
        sinusoidal_zeroes_none = sinusoidal_roots(coefficients[0], coefficients[1], coefficients[2], coefficients[3])
        self.assertEqual(sinusoidal_zeroes_none, [None])
    
    def test_sinusoidal_zeroes_many_bounce(self):
        sinusoidal_zeroes_many_bounce = sinusoidal_roots(2, 3, 5, 2)
        self.assertEqual(sinusoidal_zeroes_many_bounce, [4.4764, 6.5708, 8.6652, 10.7596, 12.854, '4.4764 + 2.0944k'])
    
    def test_sinusoidal_zeroes_many_cross(self):
        sinusoidal_zeroes_many_cross = sinusoidal_roots(2, 3, 5, 1)
        self.assertEqual(sinusoidal_zeroes_many_cross, [4.8255, 6.2217, 6.9199, 8.3161, 9.0143, 10.4105, 11.1087, 12.5049, 13.203, 14.5993, '4.8255 + 2.0944k', '6.2217 + 2.0944k'])

    def test_sinusoidal_zeroes_first_derivative(self):
        sinusoidal_zeroes_first_derivative = sinusoidal_roots_first_derivative(coefficients[0], coefficients[1], coefficients[2], coefficients[3])
        self.assertEqual(sinusoidal_zeroes_first_derivative, [5.5236, 6.5708, 7.618, 8.6652, 9.7124, '5.5236 + 1.0472k'])
    
    def test_sinusoidal_zeroes_second_derivative(self):
        sinusoidal_zeroes_second_derivative = sinusoidal_roots_second_derivative(coefficients[0], coefficients[1], coefficients[2], coefficients[3])
        self.assertEqual(sinusoidal_zeroes_second_derivative, [5.0, 6.0472, 7.0944, 8.1416, 9.1888, '5.0 + 1.0472k'])
    
    def test_sinusoidal_zeroes_initial_value_none(self):
        sinusoidal_zeroes_initial_value_none = sinusoidal_roots_initial_value(coefficients[0], coefficients[1], coefficients[2], coefficients[3], 10)
        self.assertEqual(sinusoidal_zeroes_initial_value_none, [None])

    def test_sinusoidal_zeroes_initial_value_many_bounce(self):
        sinusoidal_zeroes_initial_value_many_bounce = sinusoidal_roots_initial_value(2, 3, 5, 9, 7)
        self.assertEqual(sinusoidal_zeroes_initial_value_many_bounce, [4.4764, 6.5708, 8.6652, 10.7596, 12.854, '4.4764 + 2.0944k'])
    
    def test_sinusoidal_zeroes_initial_value_many_cross(self):
        sinusoidal_zeroes_initial_value_many_cross = sinusoidal_roots_initial_value(7, 5, 3, 4, 2)
        self.assertEqual(sinusoidal_zeroes_initial_value_many_cross, [2.942, 3.6863, 4.1987, 4.9429, 5.4553, 6.1995, 6.712, 7.4562, 7.9686, 8.7128, '2.942 + 1.2566k', '3.6863 + 1.2566k'])
    
    def test_sinusoidal_zeroes_derivative_initial_value_none(self):
        sinusoidal_zeroes_derivative_initial_value_none = sinusoidal_roots_derivative_initial_value(coefficients[0], coefficients[1], coefficients[2], coefficients[3], 10)
        self.assertEqual(sinusoidal_zeroes_derivative_initial_value_none, [None])
    
    def test_sinusoidal_zeroes_derivative_initial_value_many_bounce(self):
        sinusoidal_zeroes_derivative_initial_value_many_bounce = sinusoidal_roots_derivative_initial_value(1, 1, 2, 3, 1)
        self.assertEqual(sinusoidal_zeroes_derivative_initial_value_many_bounce, [2.0, 8.2832, 14.5664, 20.8496, 27.1327, '2.0 + 6.2832k'])
    
    def test_sinusoidal_zeroes_derivative_initial_value_many_cross(self):
        sinusoidal_zeroes_derivative_initial_value_many_cross = sinusoidal_roots_derivative_initial_value(7, 5, 3, 4, 2)
        self.assertEqual(sinusoidal_zeroes_derivative_initial_value_many_cross, [3.3027, 3.9539, 4.5594, 5.2105, 5.816, 6.4672, 7.0726, 7.7238, 8.3293, 8.9805, '3.3027 + 1.2566k', '3.9539 + 1.2566k'])

class TestIntercepts(unittest.TestCase):
    maxDiff = None

    def test_linear_intercepts(self):
        linear_intercepts = intercept_points('linear', coefficients[:2])
        self.assertEqual(linear_intercepts, [-1.5])
    
    def test_quadratic_intercepts(self):
        quadratic_intercepts = intercept_points('quadratic', coefficients[:3])
        self.assertEqual(quadratic_intercepts, [None])
    
    def test_cubic_intercepts(self):
        cubic_intercepts = intercept_points('cubic', coefficients)
        self.assertEqual(cubic_intercepts, [-1.4455])
    
    def test_hyperbolic_intercepts(self):
        hyperbolic_intercepts = intercept_points('hyperbolic', coefficients[:2])
        self.assertEqual(hyperbolic_intercepts, [-0.6667])
    
    def test_exponential_intercepts(self):
        exponential_intercepts = intercept_points('exponential', coefficients[:2])
        self.assertEqual(exponential_intercepts, [None])
    
    def test_logarithmic_intercepts(self):
        logarithmic_intercepts = intercept_points('logarithmic', coefficients[:2])
        self.assertEqual(logarithmic_intercepts, [0.2231])
    
    def test_logistic_intercepts(self):
        logistic_intercepts = intercept_points('logistic', coefficients[:3])
        self.assertEqual(logistic_intercepts, [None])
    
    def test_sinusoidal_intercepts(self):
        sinusoidal_intercepts = intercept_points('sinusoidal', coefficients)
        self.assertEqual(sinusoidal_intercepts, [None])

class TestMaxima(unittest.TestCase):
    maxDiff = None

    def test_linear_maxima(self):
        linear_maxima = maxima_points('linear', coefficients[:2])
        self.assertEqual(linear_maxima, [None])
    
    def test_quadratic_maxima(self):
        quadratic_maxima = maxima_points('quadratic', coefficients[:3])
        self.assertEqual(quadratic_maxima, [None])
    
    def test_cubic_maxima_none(self):
        cubic_maxima_none = maxima_points('cubic', coefficients[:4])
        self.assertEqual(cubic_maxima_none, [None])
    
    def test_cubic_maxima_one(self):
        cubic_maxima_one = maxima_points('cubic', [2, -15, 36, 7])
        self.assertEqual(cubic_maxima_one, [2.0])
    
    def test_hyperbolic_maxima(self):
        hyperbolic_maxima = maxima_points('hyperbolic', coefficients[:2])
        self.assertEqual(hyperbolic_maxima, [None])
    
    def test_exponential_maxima(self):
        exponential_maxima = maxima_points('exponential', coefficients[:2])
        self.assertEqual(exponential_maxima, [None])
    
    def test_logarithmic_maxima(self):
        logarithmic_maxima = maxima_points('logarithmic', coefficients[:2])
        self.assertEqual(logarithmic_maxima, [None])
    
    def test_logistic_maxima(self):
        logistic_maxima = maxima_points('logistic', coefficients[:3])
        self.assertEqual(logistic_maxima, [None])
    
    def test_sinusoidal_maxima(self):
        sinusoidal_maxima = maxima_points('sinusoidal', coefficients[:4])
        self.assertEqual(sinusoidal_maxima, [5.5236, 7.618, 9.7124])

class TestMinima(unittest.TestCase):
    maxDiff = None

    def test_linear_minima(self):
        linear_minima = minima_points('linear', coefficients[:2])
        self.assertEqual(linear_minima, [None])
    
    def test_quadratic_minima(self):
        quadratic_minima = minima_points('quadratic', coefficients[:3])
        self.assertEqual(quadratic_minima, [-0.75])
    
    def test_cubic_minima_none(self):
        cubic_minima_none = minima_points('cubic', coefficients[:4])
        self.assertEqual(cubic_minima_none, [None])
    
    def test_cubic_minima_one(self):
        cubic_minima_one = minima_points('cubic', [2, -15, 36, 7])
        self.assertEqual(cubic_minima_one, [3.0])
    
    def test_hyperbolic_minima(self):
        hyperbolic_minima = minima_points('hyperbolic', coefficients[:2])
        self.assertEqual(hyperbolic_minima, [None])
    
    def test_exponential_minima(self):
        exponential_minima = minima_points('exponential', coefficients[:2])
        self.assertEqual(exponential_minima, [None])
    
    def test_logarithmic_minima(self):
        logarithmic_minima = minima_points('logarithmic', coefficients[:2])
        self.assertEqual(logarithmic_minima, [None])
    
    def test_logistic_minima(self):
        logistic_minima = minima_points('logistic', coefficients[:3])
        self.assertEqual(logistic_minima, [None])
    
    def test_sinusoidal_minima(self):
        sinusoidal_minima = minima_points('sinusoidal', coefficients[:4])
        self.assertEqual(sinusoidal_minima, [6.5708, 8.6652])

class TestExtrema(unittest.TestCase):
    maxDiff = None

    def test_linear_extrema(self):
        linear_extrema = extrema_points('linear', coefficients[:2])
        self.assertEqual(linear_extrema, {'maxima': [None], 'minima': [None]})
    
    def test_quadratic_extrema(self):
        quadratic_extrema = extrema_points('quadratic', coefficients[:3])
        self.assertEqual(quadratic_extrema, {'maxima': [None], 'minima': [-0.75]})
    
    def test_cubic_extrema(self):
        cubic_extrema = extrema_points('cubic', coefficients[:4])
        self.assertEqual(cubic_extrema, {'maxima': [None], 'minima': [None]})
    
    def test_hyperbolic_extrema(self):
        hyperbolic_extrema = extrema_points('hyperbolic', coefficients[:2])
        self.assertEqual(hyperbolic_extrema, {'maxima': [None], 'minima': [None]})
    
    def test_exponential_extrema(self):
        exponential_extrema = extrema_points('exponential', coefficients[:2])
        self.assertEqual(exponential_extrema, {'maxima': [None], 'minima': [None]})
    
    def test_logarithmic_extrema(self):
        logarithmic_extrema = extrema_points('logarithmic', coefficients[:2])
        self.assertEqual(logarithmic_extrema, {'maxima': [None], 'minima': [None]})
    
    def test_logistic_extrema(self):
        logistic_extrema = extrema_points('logistic', coefficients[:3])
        self.assertEqual(logistic_extrema, {'maxima': [None], 'minima': [None]})
    
    def test_sinusoidal_extrema(self):
        sinusoidal_extrema = extrema_points('sinusoidal', coefficients[:4])
        self.assertEqual(sinusoidal_extrema, {'maxima': [5.5236, 7.618, 9.7124, '5.5236 + 2.0944k'], 'minima': [6.5708, 8.6652, '6.5708 + 2.0944k']})

class TestInflections(unittest.TestCase):
    maxDiff = None

    def test_linear_inflections(self):
        linear_inflections = inflection_points('linear', coefficients[:2])
        self.assertEqual(linear_inflections, [None])
    
    def test_quadratic_inflections(self):
        quadratic_inflections = inflection_points('quadratic', coefficients[:3])
        self.assertEqual(quadratic_inflections, [None])
    
    def test_cubic_inflections(self):
        cubic_inflections = inflection_points('cubic', coefficients[:4])
        self.assertEqual(cubic_inflections, [-0.5])
    
    def test_hyperbolic_inflections(self):
        hyperbolic_inflections = inflection_points('hyperbolic', coefficients[:2])
        self.assertEqual(hyperbolic_inflections, [None])
    
    def test_exponential_inflections(self):
        exponential_inflections = inflection_points('exponential', coefficients[:2])
        self.assertEqual(exponential_inflections, [None])
    
    def test_logarithmic_inflections(self):
        logarithmic_inflections = inflection_points('logarithmic', coefficients[:2])
        self.assertEqual(logarithmic_inflections, [None])
    
    def test_logistic_inflections(self):
        logistic_inflections = inflection_points('logistic', coefficients[:3])
        self.assertEqual(logistic_inflections, [5])
    
    def test_sinusoidal_inflections(self):
        sinusoidal_inflections = inflection_points('sinusoidal', coefficients[:4])
        self.assertEqual(sinusoidal_inflections, [5.0, 6.0472, 7.0944, 8.1416, 9.1888, '5.0 + 1.0472k'])

class TestCoordinatePairs(unittest.TestCase):
    maxDiff = None

    def test_coordinate_pairs_linear(self):
        coordinate_pairs_linear = coordinate_pairs('linear', coefficients[:2], [1, 2, 3, 4])
        self.assertEqual(coordinate_pairs_linear, [[1.0, 5.0], [2.0, 7.0], [3.0, 9.0], [4.0, 11.0]])
    
    def test_coordinate_pairs_quadratic(self):
        coordinate_pairs_quadratic = coordinate_pairs('quadratic', coefficients[:3], [1, 2, 3, 4])
        self.assertEqual(coordinate_pairs_quadratic, [[1.0, 10.0], [2.0, 19.0], [3.0, 32.0], [4.0, 49.0]])
    
    def test_coordinate_pairs_cubic(self):
        coordinate_pairs_cubic = coordinate_pairs('cubic', coefficients[:4], [1, 2, 3, 4])
        self.assertEqual(coordinate_pairs_cubic, [[1.0, 17.0], [2.0, 45.0], [3.0, 103.0], [4.0, 203.0]])
    
    def test_coordinate_pairs_hyperbolic(self):
        coordinate_pairs_hyperbolic = coordinate_pairs('hyperbolic', coefficients[:2], [1, 2, 3, 4])
        self.assertEqual(coordinate_pairs_hyperbolic, [[1.0, 5.0], [2.0, 4.0], [3.0, 3.6667], [4.0, 3.5]])
    
    def test_coordinate_pairs_exponential(self):
        coordinate_pairs_exponential = coordinate_pairs('exponential', coefficients[:2], [1, 2, 3, 4])
        self.assertEqual(coordinate_pairs_exponential, [[1.0, 6.0], [2.0, 18.0], [3.0, 54.0], [4.0, 162.0]])
    
    def test_coordinate_pairs_logarithmic(self):
        coordinate_pairs_logarithmic = coordinate_pairs('logarithmic', coefficients[:2], [1, 2, 3, 4])
        self.assertEqual(coordinate_pairs_logarithmic, [[1.0, 3.0], [2.0, 4.3863], [3.0, 5.1972], [4.0, 5.7726]])
    
    def test_coordinate_pairs_logistic(self):
        coordinate_pairs_logistic = coordinate_pairs('logistic', coefficients[:3], [1, 2, 3, 4])
        self.assertEqual(coordinate_pairs_logistic, [[1.0, 0.0001], [2.0, 0.0002], [3.0, 0.0049], [4.0, 0.0949]])
    
    def test_coordinate_pairs_sinusoidal(self):
        coordinate_pairs_sinusoidal = coordinate_pairs('sinusoidal', coefficients[:4], [1, 2, 3, 4])
        self.assertEqual(coordinate_pairs_sinusoidal, [[1.0, 8.0731], [2.0, 6.1758], [3.0, 7.5588], [4.0, 6.7178]])
    
    def test_coordinate_pairs_sinusoidal_mixed(self):
        coordinate_pairs_sinusoidal_mixed = coordinate_pairs('sinusoidal', coefficients[:4], [1, 2, 3, 4, '1 + 1k'])
        self.assertEqual(coordinate_pairs_sinusoidal_mixed,  [[1.0, 8.0731], [2.0, 6.1758], [3.0, 7.5588], [4.0, 6.7178], ['1 + 1k', 8.0731]])
    
    def test_coordinate_pairs_sinusoidal_mixed_intercepts(self):
        coordinate_pairs_sinusoidal_mixed_intercepts = coordinate_pairs('sinusoidal', coefficients[:4], [1, 2, 3, 4, '1 + 1k'], 'intercepts')
        self.assertEqual(coordinate_pairs_sinusoidal_mixed_intercepts, [[1.0, 0.0], [2.0, 0.0], [3.0, 0.0], [4.0, 0.0], ['1 + 1k', 0.0]])
    
    def test_coordinate_pairs_sinusoidal_mixed_inflections(self):
        coordinate_pairs_sinusoidal_mixed_inflections = coordinate_pairs('sinusoidal', coefficients[:4], [1, 2, 3, 4, '1 + 1k'], 'inflections')
        self.assertEqual(coordinate_pairs_sinusoidal_mixed_inflections,  [[1.0, 8.0731], [2.0, 6.1758], [3.0, 7.5588], [4.0, 6.7178], ['1 + 1k', 8.0731]])
    
    def test_coordinate_pairs_sinusoidal_mixed_maxima(self):
        coordinate_pairs_sinusoidal_mixed_maxima = coordinate_pairs('sinusoidal', coefficients[:4], [1, 2, 3, 4, '1 + 1k'], 'maxima')
        self.assertEqual(coordinate_pairs_sinusoidal_mixed_maxima, [[1.0, 8.0731], [2.0, 6.1758], [3.0, 7.5588], [4.0, 6.7178], ['1 + 1k', 8.0731]])
    
    def test_coordinate_pairs_sinusoidal_mixed_minima(self):
        coordinate_pairs_sinusoidal_mixed_minima = coordinate_pairs('sinusoidal', coefficients[:4], [1, 2, 3, 4, '1 + 1k'], 'minima')
        self.assertEqual(coordinate_pairs_sinusoidal_mixed_minima, [[1.0, 8.0731], [2.0, 6.1758], [3.0, 7.5588], [4.0, 6.7178], ['1 + 1k', 8.0731]])

class TestKeyPoints(unittest.TestCase):
    maxDiff = None

    def test_linear_key_points(self):
        linear_key_points = key_coordinates('linear', coefficients[:2])
        self.assertEqual(linear_key_points, {'roots': [[-1.5, 0]], 'maxima': [None], 'minima': [None], 'inflections': [None]})
    
    def test_quadratic_key_points(self):
        quadratic_key_points = key_coordinates('quadratic', coefficients[:3])
        self.assertEqual(quadratic_key_points, {'roots': [None], 'maxima': [None], 'minima': [[-0.75, 3.875]], 'inflections': [None]})
    
    def test_cubic_key_points(self):
        cubic_key_points = key_coordinates('cubic', coefficients[:4])
        self.assertEqual(cubic_key_points, {'roots': [[-1.4455, 0]], 'maxima': [None], 'minima': [None], 'inflections': [[-0.5, 5.0]]})
    
    def test_hyperbolic_key_points(self):
        hyperbolic_key_points = key_coordinates('hyperbolic', coefficients[:2])
        self.assertEqual(hyperbolic_key_points, {'roots': [[-0.6667, 0]], 'maxima': [None], 'minima': [None], 'inflections': [None]})
    
    def test_exponential_key_points(self):
        exponential_key_points = key_coordinates('exponential', coefficients[:2])
        self.assertEqual(exponential_key_points, {'roots': [None], 'maxima': [None], 'minima': [None], 'inflections': [None]})
    
    def test_logarithmic_key_points(self):
        logarithmic_key_points = key_coordinates('logarithmic', coefficients[:2])
        self.assertEqual(logarithmic_key_points, {'roots': [[0.2231, 0]], 'maxima': [None], 'minima': [None], 'inflections': [None]})
    
    def test_logistic_key_points(self):
        logistic_key_points = key_coordinates('logistic', coefficients[:3])
        self.assertEqual(logistic_key_points, {'roots': [None], 'maxima': [None], 'minima': [None], 'inflections': [[5, 1.0]]})
    
    def test_sinusoidal_key_points(self):
        sinusoidal_key_points = key_coordinates('sinusoidal', coefficients[:4])
        self.assertEqual(sinusoidal_key_points, {'roots': [None], 'maxima': [[5.5236, 9.0], [7.618, 9.0], [9.7124, 9.0], ['5.5236 + 2.0944k', 9.0]], 'minima': [[6.5708, 5.0], [8.6652, 5.0], ['6.5708 + 2.0944k', 5.0]], 'inflections': [[5.0, 7.0], [6.0472, 7.0], [7.0944, 7.0], [8.1416, 7.0], [9.1888, 7.0001], ['5.0 + 1.0472k', 7.0]]})

class TestPointsWithinRange(unittest.TestCase):
    maxDiff = None

    def test_points_range_all(self):
        points_range_all = points_within_range([11, 13, 15, 17, 19], 11, 19)
        self.assertEqual(points_range_all, [11, 13, 15, 17, 19])
    
    def test_points_range_some(self):
        points_range_some = points_within_range([11, 13, 15, 17, 19], 14, 18)
        self.assertEqual(points_range_some, [15, 17])
    
    def test_points_range_none(self):
        points_range_none = points_within_range([11, 13, 15, 17, 19], 50, 60)
        self.assertEqual(points_range_none, [None])

class TestShiftedPointsWithinRange(unittest.TestCase):
    maxDiff = None

    def test_shifted_points_in_range_1d(self):
        shifted_points_in_range_1d = shifted_points_within_range([11, 13, 15, 17, 19, '1 + 2k'], 11, 19)
        self.assertEqual(shifted_points_in_range_1d, [11.0, 13.0, 15.0, 17.0, 19.0, '11.0 + 2.0k'])
    
    def test_shifted_points_not_in_range_1d(self):
        shifted_points_not_in_range_1d = shifted_points_within_range([11, 13, 15, 17, 19, '1 + 2k'], 50, 60)
        self.assertEqual(shifted_points_not_in_range_1d, [51.0, 53.0, 55.0, 57.0, 59.0, '51.0 + 2.0k'])
    
    def test_shifted_points_in_range_2d(self):
        shifted_points_in_range_2d = shifted_points_within_range([[11, 1], [13, 1], [15, 1], [17, 1], [19, 1], ['1 + 2k', 1]], 11, 19)
        self.assertEqual(shifted_points_in_range_2d, [11.0, 13.0, 15.0, 17.0, 19.0, '11.0 + 2.0k'])
    
    def test_shifted_points_not_in_range_2d(self):
        shifted_points_not_in_range_2d = shifted_points_within_range([[11, 1], [13, 1], [15, 1], [17, 1], [19, 1], ['1 + 2k', 1]], 50, 60)
        self.assertEqual(shifted_points_not_in_range_2d, [51.0, 53.0, 55.0, 57.0, 59.0, '51.0 + 2.0k'])

class TestShiftedCoordinatesWithinRange(unittest.TestCase):
    maxDiff = None

    def test_shifted_coordinates_in_range(self):
        shifted_coordinates_in_range = shifted_coordinates_within_range([[11, 1], [13, 1], [15, 1], [17, 1], [19, 1], ['1 + 2k', 1]], 11, 19, 8)
        self.assertEqual(shifted_coordinates_in_range, [[11.0, 1], [13.0, 1], [15.0, 1], [17.0, 1], [19.0, 1], ['11.0 + 2.0k', 1]])
    
    def test_shifted_coordinates_not_in_range(self):
        shifted_coordinates_not_in_range = shifted_coordinates_within_range([[11, 1], [13, 1], [15, 1], [17, 1], [19, 1], ['1 + 2k', 1]], 50, 60, 5)
        self.assertEqual(shifted_coordinates_not_in_range, [[51.0, 1], [53.0, 1], [55.0, 1], [57.0, 1], [59.0, 1], ['51.0 + 2.0k', 1]])
    
    def test_shifted_coordinates_none(self):
        shifted_coordinates_none = shifted_coordinates_within_range([None], 50, 60, 5)
        self.assertEqual(shifted_coordinates_none, [None])

class TestAccumulation(unittest.TestCase):
    def test_linear_accumulation(self):
        linear_accumulation = accumulated_area('linear', coefficients[:2], 10, 20)
        self.assertEqual(linear_accumulation, 330.0)
    
    def test_quadratic_accumulation(self):
        quadratic_accumulation = accumulated_area('quadratic', coefficients[:3], 10, 20)
        self.assertEqual(quadratic_accumulation, 5166.9)
    
    def test_cubic_accumulation(self):
        cubic_accumulation = accumulated_area('cubic', coefficients[:4], 10, 20)
        self.assertEqual(cubic_accumulation, 82820.0)
    
    def test_hyperbolic_accumulation(self):
        hyperbolic_accumulation = accumulated_area('hyperbolic', coefficients[:2], 10, 20)
        self.assertEqual(hyperbolic_accumulation, 31.3863)
    
    def test_exponential_accumulation(self):
        exponential_accumulation = accumulated_area('exponential', coefficients[:2], 10, 20)
        self.assertEqual(exponential_accumulation, 6347583503.316)
    
    def test_logarithmic_accumulation(self):
        logarithmic_accumulation = accumulated_area('logarithmic', coefficients[:2], 10, 20)
        self.assertEqual(logarithmic_accumulation, 83.7776)
    
    def test_logistic_accumulation(self):
        logistic_accumulation = accumulated_area('logistic', coefficients[:3], 10, 20)
        self.assertEqual(logistic_accumulation, 20.001)
    
    def test_sinusoidal_accumulation(self):
        sinusoidal_accumulation = accumulated_area('sinusoidal', coefficients[:4], 10, 20)
        self.assertEqual(sinusoidal_accumulation, 69.1433)

class TestAverages(unittest.TestCase):
    maxDiff = None

    def test_linear_averages(self):
        linear_averages = average_values('linear', coefficients[:2], 10, 20)
        self.assertEqual(linear_averages, {'average_value_derivative': 2.0, 'mean_values_derivative': ['All'], 'average_value_integral': 33.0, 'mean_values_integral': [15.0]})
    
    def test_quadratic_averages(self):
        quadratic_averages = average_values('quadratic', coefficients[:3], 10, 20)
        self.assertEqual(quadratic_averages, {'average_value_derivative': 63.0, 'mean_values_derivative': [15.0], 'average_value_integral': 516.69, 'mean_values_integral': [15.2627]})
    
    def test_cubic_averages(self):
        cubic_averages = average_values('cubic', coefficients[:4], 10, 20)
        self.assertEqual(cubic_averages, {'average_value_derivative': 1495.0, 'mean_values_derivative': [15.2665], 'average_value_integral': 8282.0, 'mean_values_integral': [15.5188]})
    
    def test_hyperbolic_averages(self):
        hyperbolic_averages = average_values('hyperbolic', coefficients[:2], 10, 20)
        self.assertEqual(hyperbolic_averages, {'average_value_derivative': -0.01, 'mean_values_derivative': [14.1421], 'average_value_integral': 3.1386, 'mean_values_integral': [14.43]})
    
    def test_exponential_averages(self):
        exponential_averages = average_values('exponential', coefficients[:2], 10, 20)
        self.assertEqual(exponential_averages, {'average_value_derivative': 697345070.4, 'mean_values_derivative': [17.8185], 'average_value_integral': 634758350.3316, 'mean_values_integral': [17.8185]})
    
    def test_exponential_averages_unary_base(self):
        exponential_averages_unary_base = average_values('exponential', [2, 1], 10, 20)
        self.assertEqual(exponential_averages_unary_base, {'average_value_derivative': 0.0, 'mean_values_derivative': [None], 'average_value_integral': 2.003, 'mean_values_integral': [14.9888]})
    
    def test_exponential_averages_negative_base(self):
        exponential_averages_negative_base = average_values('exponential', [2, -3], 10, 20)
        self.assertEqual(exponential_averages_negative_base, {'average_value_derivative': 697345070.4, 'mean_values_derivative': [17.8185], 'average_value_integral': 634758350.3316, 'mean_values_integral': [17.8185]})
    
    def test_exponential_averages_negative_lead(self):
        exponential_averages_negative_lead = average_values('exponential', [-2, 3], 10, 20)
        self.assertEqual(exponential_averages_negative_lead, {'average_value_derivative': -697345070.4, 'mean_values_derivative': [17.8185], 'average_value_integral': -634758350.3316, 'mean_values_integral': [17.8185]})
    
    def test_logarithmic_averages(self):
        logarithmic_averages = average_values('logarithmic', coefficients[:2], 10, 20)
        self.assertEqual(logarithmic_averages, {'average_value_derivative': 0.1386, 'mean_values_derivative': [14.43], 'average_value_integral': 8.3778, 'mean_values_integral': [14.7155]})
    
    def test_logistic_averages(self):
        logistic_averages = average_values('logistic', coefficients[:3], 10, 20)
        self.assertEqual(logistic_averages, {'average_value_derivative': 0.0, 'mean_values_derivative': [None], 'average_value_integral': 2.0001, 'mean_values_integral': [None]})
    
    def test_sinusoidal_averages(self):
        sinusoidal_averages = average_values('sinusoidal', coefficients[:4], 10, 20)
        self.assertEqual(sinusoidal_averages, {'average_value_derivative': 0.0401, 'mean_values_derivative': [10.7618, 11.8046, 12.8562, 13.899, 14.9506, 15.9934, 17.045, 18.0878, 19.1394, '10.7618 + 2.0944k', '11.8046 + 2.0944k'], 'average_value_integral': 6.9143, 'mean_values_integral': [10.2503, 11.2689, 12.3447, 13.3633, 14.4391, 15.4577, 16.5335, 17.5521, 18.6279, 19.6465, '10.2503 + 2.0944k', '11.2689 + 2.0944k']})
    
    def test_averages_start_end(self):
        averages_start_end = average_values('cubic', coefficients[:4], 1, 1)
        self.assertEqual(averages_start_end, {'average_value_derivative': 0.0, 'mean_values_derivative': [None], 'average_value_integral': 0.0, 'mean_values_integral': [None]})

if __name__ == '__main__':
    unittest.main()

# ---------- Ran 322 tests in 0.049s ---------- OK ---------- #


import unittest

from library.matrices.addition import matrix_sum
from library.matrices.multiplication import scalar_product_matrix, matrix_product
from library.matrices.transpose import transposed_matrix
from library.matrices.cofactors import matrix_of_cofactors
from library.matrices.determinant import linear_determinant
from library.matrices.minors import matrix_of_minors
from library.matrices.inverse import inverse_matrix
from library.matrices.solve import system_solution

first_2d = [
    [5, 8],
    [2, 3]
]

second_2d = [
    [4, 1],
    [7, 3]
]

zero_2d = [
    [0, 0],
    [0, 0]
]

first_3d = [
    [6, 1, 1],
    [4, -2, 5],
    [2, 8, 7]
]

second_3d = [
    [3, 0, 2],
    [2, 0, -2],
    [0, 1, 1]
]

first_4d = [
    [5, 2, -6, 2],
    [3, 4, 1, -5],
    [9, -8, 7, 1],
    [-2, 5, 0, 11]
]

second_4d = [
    [13, -7, 1, -12],
    [5, -16, 14, 5],
    [-8, 6, 2, 11],
    [3, 0, 8, -2]
]

first_2x3 = [
    [2, 6, -9],
    [4, 5, 1]
]

second_2x3 = [
    [7, 1, 3],
    [-8, 2, 5]
]

column_2d = [
    [3],
    [-4]
]

column_3d = [
    [5],
    [-1],
    [7]
]

scalar_number = -7

class TestAddition(unittest.TestCase):
    def test_addition_2d(self):
        addition_2d = matrix_sum(first_2d, second_2d)
        self.assertEqual(addition_2d, [[9, 9], [9, 6]])
    
    def test_addition_3d(self):
        addition_3d = matrix_sum(first_3d, second_3d)
        self.assertEqual(addition_3d, [[9, 1, 3], [6, -2, 3], [2, 9, 8]])
    
    def test_addition_4d(self):
        addition_4d = matrix_sum(first_4d, second_4d)
        self.assertEqual(addition_4d, [[18, -5, -5, -10], [8, -12, 15, 0], [1, -2, 9, 12], [1, 5, 8, 9]])
    
    def test_addition_2x3(self):
        addition_2x3 = matrix_sum(first_2x3, second_2x3)
        self.assertEqual(addition_2x3, [[9, 7, -6], [-4, 7, 6]])

class TestScalarProduct(unittest.TestCase):
    def test_scalar_2d(self):
        scalar_2d = scalar_product_matrix(first_2d, scalar_number)
        self.assertEqual(scalar_2d, [[-35, -56], [-14, -21]])
    
    def test_scalar_3d(self):
        scalar_3d = scalar_product_matrix(first_3d, scalar_number)
        self.assertEqual(scalar_3d, [[-42, -7, -7], [-28, 14, -35], [-14, -56, -49]])
    
    def test_scalar_4d(self):
        scalar_4d = scalar_product_matrix(first_4d, scalar_number)
        self.assertEqual(scalar_4d, [[-35, -14, 42, -14], [-21, -28, -7, 35], [-63, 56, -49, -7], [14, -35, 0, -77]])
    
    def test_scalar_2x3(self):
        scalar_2x3 = scalar_product_matrix(first_2x3, scalar_number)
        self.assertEqual(scalar_2x3, [[-14, -42, 63], [-28, -35, -7]])

class TestMatrixProduct(unittest.TestCase):
    def test_multiplication_2d_first(self):
        multiplication_2d_first = matrix_product(first_2d, second_2d)
        self.assertEqual(multiplication_2d_first, [[76, 29], [29, 11]])
    
    def test_multiplication_2d_second(self):
        multiplication_2d_second = matrix_product(second_2d, first_2d)
        self.assertEqual(multiplication_2d_second, [[22, 35], [41, 65]])
    
    def test_multiplication_3d_first(self):
        multiplication_3d_first = matrix_product(first_3d, second_3d)
        self.assertEqual(multiplication_3d_first, [[20, 1, 11], [8, 5, 17], [22, 7, -5]])
    
    def test_multiplication_3d_second(self):
        multiplication_3d_second = matrix_product(second_3d, first_3d)
        self.assertEqual(multiplication_3d_second, [[22, 19, 17], [8, -14, -12], [6, 6, 12]])
    
    def test_multiplication_4d_first(self):
        multiplication_4d_first = matrix_product(first_4d, second_4d)
        self.assertEqual(multiplication_4d_first, [[129, -103, 37, -120], [36, -79, 21, 5], [24, 107, -81, -73], [32, -66, 156, 27]])
    
    def test_multiplication_4d_second(self):
        multiplication_4d_second = matrix_product(second_4d, first_4d)
        self.assertEqual(multiplication_4d_second, [[77, -70, -78, -70], [93, -141, 52, 159], [-26, 47, 68, 77], [91, -68, 38, -8]])

class TestTranspose(unittest.TestCase):
    def test_transpose_2d(self):
        transpose_2d = transposed_matrix(first_2d)
        self.assertEqual(transpose_2d, [[5, 2], [8, 3]])
    
    def test_transpose_3d(self):
        transpose_3d = transposed_matrix(first_3d)
        self.assertEqual(transpose_3d, [[6, 4, 2], [1, -2, 8], [1, 5, 7]])
    
    def test_transpose_4d(self):
        transpose_4d = transposed_matrix(first_4d)
        self.assertEqual(transpose_4d, [[5, 3, 9, -2], [2, 4, -8, 5], [-6, 1, 7, 0], [2, -5, 1, 11]])
    
    def test_transpose_2x3(self):
        transpose_2x3 = transposed_matrix(first_2x3)
        self.assertEqual(transpose_2x3, [[2, 4], [6, 5], [-9, 1]])

class TestCofactors(unittest.TestCase):
    def test_cofactors_2d(self):
        cofactors_2d = matrix_of_cofactors(first_2d)
        self.assertEqual(cofactors_2d, [[5, -8], [-2, 3]])
    
    def test_cofactors_3d(self):
        cofactors_3d = matrix_of_cofactors(first_3d)
        self.assertEqual(cofactors_3d, [[6, -1, 1], [-4, -2, -5], [2, -8, 7]])
    
    def test_cofactors_4d(self):
        cofactors_4d = matrix_of_cofactors(first_4d)
        self.assertEqual(cofactors_4d, [[5, -2, -6, -2], [-3, 4, -1, -5], [9, 8, 7, -1], [2, 5, 0, 11]])

class TestDeterminant(unittest.TestCase):
    def test_determinant_2d(self):
        determinant_2d = linear_determinant(first_2d)
        self.assertEqual(determinant_2d, -1)
    
    def test_determinant_3d(self):
        determinant_3d = linear_determinant(first_3d)
        self.assertEqual(determinant_3d, -306)
    
    def test_determinant_4d(self):
        determinant_4d = linear_determinant(first_4d)
        self.assertEqual(determinant_4d, 7992)

class TestMinors(unittest.TestCase):
    def test_minors_2d(self):
        minors_2d = matrix_of_minors(first_2d)
        self.assertEqual(minors_2d, [[3, 2], [8, 5]])
    
    def test_minors_3d(self):
        minors_3d = matrix_of_minors(first_3d)
        self.assertEqual(minors_3d, [[-54, 18, 36], [-1, 40, 46], [7, 26, -16]])
    
    def test_minors_4d(self):
        minors_4d = matrix_of_minors(first_4d)
        self.assertEqual(minors_4d, [[576, 60, -828, -132], [-474, 1019, -609, -377], [426, 197, 345, -167], [-72, 492, -396, 516]])

class TestInverse(unittest.TestCase):
    def test_inverse_2d(self):
        inverse_2d = inverse_matrix(first_2d)
        self.assertEqual(inverse_2d, [[-3.0, 8.0], [2.0, -5.0]])
    
    def test_inverse_3d(self):
        inverse_3d = inverse_matrix(first_3d)
        self.assertEqual(inverse_3d, [[0.17647058823529413, -0.0032679738562091504, -0.022875816993464054], [0.058823529411764705, -0.13071895424836602, 0.08496732026143791], [-0.11764705882352941, 0.1503267973856209, 0.05228758169934641]])
    
    def test_inverse_4d(self):
        inverse_4d = inverse_matrix(first_4d)
        self.assertEqual(inverse_4d, [[0.07207207207207207, 0.05930930930930931, 0.0533033033033033, 0.009009009009009009], [-0.0075075075075075074, 0.1275025025025025, -0.02464964964964965, 0.06156156156156156], [-0.1036036036036036, 0.0762012012012012, 0.04316816816816817, 0.04954954954954955], [0.016516516516516516, -0.04717217217217217, 0.020895895895895897, 0.06456456456456457]])
    
    def test_inverse_zero(self):
        inverse_zero = inverse_matrix(zero_2d)
        self.assertEqual(inverse_zero, zero_2d)

class TestSolveSystems(unittest.TestCase):
    def test_solve_2d(self):
        solve_2d = system_solution(first_2d, column_2d)
        self.assertEqual(solve_2d, [-41.0, 26.0])
    
    def test_solve_3d(self):
        solve_3d = system_solution(first_3d, column_3d)
        self.assertEqual(solve_3d, [0.7255, 1.0196, -0.3725])

if __name__ == '__main__':
    unittest.main()

# ---------- Ran 33 tests in 0.012s ---------- OK ---------- #


import unittest

from library.errors.positions import argument_position
from library.errors.adjustments import no_zeroes
from library.errors.scalars import scalar_value, two_scalars, three_scalars, four_scalars, five_scalars, compare_scalars, positive_integer, whole_number, select_integers, allow_none_scalar
from library.errors.vectors import confirm_vector, vector_of_scalars, vector_of_strings, compare_vectors, allow_none_vector, length, long_vector
from library.errors.matrices import confirm_matrix, matrix_of_scalars, square_matrix, compare_rows, compare_columns, compare_matrices, columns_rows, allow_none_matrix, allow_vector_matrix, level
from library.errors.analyses import select_equations, select_points

class TestPosition(unittest.TestCase):
    def test_position_none(self):
        position_none = argument_position()
        self.assertEqual(position_none, 'argument')
    
    def test_position_only(self):
        position_only = argument_position('only')
        self.assertEqual(position_only, 'argument')
    
    def test_position_first(self):
        position_first = argument_position('first')
        self.assertEqual(position_first, 'first argument')
    
    def test_position_second(self):
        position_second = argument_position('second')
        self.assertEqual(position_second, 'second argument')

no_zeroes_list = [1, 2, 3]
all_zeroes_list = [0, 0, 0]
first_zero_list = [0, 1, 2]
last_zero_list = [1, 2, 0]

class TestNoZeroes(unittest.TestCase):
    def test_no_zeroes_no0(self):
        no_zeroes_no0 = no_zeroes(no_zeroes_list)
        self.assertEqual(no_zeroes_no0, [1.0, 2.0, 3.0])
    
    def test_no_zeroes_all0(self):
        no_zeroes_all0 = no_zeroes(all_zeroes_list)
        self.assertEqual(no_zeroes_all0, [0.0001, 0.0001, 0.0001])
    
    def test_no_zeroes_first0(self):
        no_zeroes_first0 = no_zeroes(first_zero_list)
        self.assertEqual(no_zeroes_first0, [0.0001, 1.0, 2.0])
    
    def test_no_zeroes_last0(self):
        no_zeroes_last0 = no_zeroes(last_zero_list)
        self.assertEqual(no_zeroes_last0, [1.0, 2.0, 0.0001])
    
    def test_no_zeroes_all0_precision(self):
        no_zeroes_all0 = no_zeroes(all_zeroes_list, 6)
        self.assertEqual(no_zeroes_all0, [0.000001, 0.000001, 0.000001])

good_positive = 3
good_whole = 0
good_integer = -3
good_float = 3.14
bad_scalar = 'three'
none_scalar = None
choices = [4, 5, 6]

class TestScalarValue(unittest.TestCase):
    def test_scalar_integer(self):
        scalar_integer = scalar_value(good_integer)
        self.assertEqual(scalar_integer, 'Argument is an integer or a float')
    
    def test_scalar_float(self):
        scalar_float = scalar_value(good_float)
        self.assertEqual(scalar_float, 'Argument is an integer or a float')
    
    def test_scalar_position(self):
        scalar_position = scalar_value(good_float, 'second')
        self.assertEqual(scalar_position, 'Second argument is an integer or a float')
    
    def test_scalar_string_raises(self):
        with self.assertRaises(Exception) as context:
            scalar_value(bad_scalar)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Argument must be an integer or a float')
    
    def test_scalar_none_raises(self):
        with self.assertRaises(Exception) as context:
            scalar_value(none_scalar)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Argument must be an integer or a float')
    
    def test_scalar_array_raises(self):
        with self.assertRaises(Exception) as context:
            scalar_value(choices)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Argument must be an integer or a float')

class TestWholeNumber(unittest.TestCase):
    def test_whole_zero(self):
        whole_zero = whole_number(good_whole, 'second')
        self.assertEqual(whole_zero, 'Second argument is a whole number')
    
    def test_whole_natural(self):
        whole_natural = whole_number(good_positive, 'third')
        self.assertEqual(whole_natural, 'Third argument is a whole number')
    
    def test_whole_negative_raises(self):
        with self.assertRaises(Exception) as context:
            whole_number(good_integer, 'second')
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'Second argument must be a whole number')
    
    def test_whole_float_raises(self):
        with self.assertRaises(Exception) as context:
            whole_number(good_float, 'second')
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'Second argument must be a whole number')
    
    def test_whole_string_raises(self):
        with self.assertRaises(Exception) as context:
            whole_number(bad_scalar, 'second')
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'Second argument must be a whole number')
    
    def test_whole_none_raises(self):
        with self.assertRaises(Exception) as context:
            whole_number(none_scalar, 'second')
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'Second argument must be a whole number')
    
    def test_whole_array_raises(self):
        with self.assertRaises(Exception) as context:
            whole_number(choices, 'second')
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'Second argument must be a whole number')

class TestPositiveInteger(unittest.TestCase):
    def test_positive_natural(self):
        positive_natural = positive_integer(good_positive)
        self.assertEqual(positive_natural, 'Last argument is a positive integer')
    
    def test_positive_whole_raises(self):
        with self.assertRaises(Exception) as context:
            positive_integer(good_whole)
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'Last argument must be a positive integer')
    
    def test_positive_negative_raises(self):
        with self.assertRaises(Exception) as context:
            positive_integer(good_integer)
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'Last argument must be a positive integer')
    
    def test_positive_float_raises(self):
        with self.assertRaises(Exception) as context:
            positive_integer(good_float)
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'Last argument must be a positive integer')
    
    def test_positive_string_raises(self):
        with self.assertRaises(Exception) as context:
            positive_integer(bad_scalar)
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'Last argument must be a positive integer')
    
    def test_positive_none_raises(self):
        with self.assertRaises(Exception) as context:
            positive_integer(none_scalar)
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'Last argument must be a positive integer')
    
    def test_positive_array_raises(self):
        with self.assertRaises(Exception) as context:
            positive_integer(choices)
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'Last argument must be a positive integer')

class TestAllowNoneScalar(unittest.TestCase):
    def test_none_scalar_none(self):
        none_scalar_none = allow_none_scalar(none_scalar)
        self.assertEqual(none_scalar_none, 'Argument is an integer, a float, or None')
    
    def test_none_scalar_integer(self):
        none_scalar_integer = allow_none_scalar(good_integer)
        self.assertEqual(none_scalar_integer, 'Argument is an integer, a float, or None')
    
    def test_none_scalar_float(self):
        none_scalar_float = allow_none_scalar(good_float)
        self.assertEqual(none_scalar_float, 'Argument is an integer, a float, or None')
    
    def test_none_scalar_string_raises(self):
        with self.assertRaises(Exception) as context:
            allow_none_scalar(bad_scalar)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Argument must be an integer, a float, or None')
    
    def test_none_scalar_array_raises(self):
        with self.assertRaises(Exception) as context:
            allow_none_scalar(choices)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Argument must be an integer, a float, or None')

class TestCompareScalars(unittest.TestCase):
    def test_compare_scalars_less(self):
        compare_scalars_less = compare_scalars(good_integer, good_float, 'second', 'third')
        self.assertEqual(compare_scalars_less, 'Second argument is less than or equal to third argument')
    
    def test_compare_scalars_equal(self):
        compare_scalars_equal = compare_scalars(good_integer, good_integer, 'second', 'third')
        self.assertEqual(compare_scalars_equal, 'Second argument is less than or equal to third argument')
    
    def test_compare_scalars_more_raises(self):
        with self.assertRaises(Exception) as context:
            compare_scalars(good_float, good_integer, 'second', 'third')
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'Second argument must be less than or equal to third argument')

class TestSelectIntegers(unittest.TestCase):
    def test_select_integers_included(self):
        select_integers_included = select_integers(5, choices)
        self.assertEqual(select_integers_included, 'Argument is one of the following integers: [4, 5, 6]')
    
    def test_select_integers_excluded_raises(self):
        with self.assertRaises(Exception) as context:
            select_integers(1, choices)
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'Argument must be one of the following integers: [4, 5, 6]')

class TestTwoScalars(unittest.TestCase):
    def test_two_scalars_integer_float(self):
        two_scalars_integer_float = two_scalars(good_integer, good_float)
        self.assertEqual(two_scalars_integer_float, 'Both first and second arguments are integers or floats')
    
    def test_two_scalars_string_integer_raises(self):
        with self.assertRaises(Exception) as context:
            two_scalars(bad_scalar, good_integer)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'First argument must be an integer or a float')
    
    def test_two_scalars_integer_string_raises(self):
        with self.assertRaises(Exception) as context:
            two_scalars(good_integer, bad_scalar)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Second argument must be an integer or a float')

class TestThreeScalars(unittest.TestCase):
    def test_three_scalars_integer_float_whole(self):
        three_scalars_integer_float_whole = three_scalars(good_integer, good_float, good_whole)
        self.assertEqual(three_scalars_integer_float_whole, 'First, second, and third arguments are all integers or floats')
    
    def test_three_scalars_integer_float_string_raises(self):
        with self.assertRaises(Exception) as context:
            three_scalars(good_integer, good_float, bad_scalar)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Third argument must be an integer or a float')

class TestFourScalars(unittest.TestCase):
    def test_four_scalars_integer_float_whole_positive(self):
        four_scalars_integer_float_whole_positive = four_scalars(good_integer, good_float, good_whole, good_positive)
        self.assertEqual(four_scalars_integer_float_whole_positive, 'First, second, third, and fourth arguments are all integers or floats')
    
    def test_four_scalars_integer_float_whole_string_raises(self):
        with self.assertRaises(Exception) as context:
            four_scalars(good_integer, good_float, good_whole, bad_scalar)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Fourth argument must be an integer or a float')

class TestFiveScalars(unittest.TestCase):
    def test_five_scalars_integer_float_whole_positive(self):
        five_scalars_integer_float_whole_positive = five_scalars(good_integer, good_float, good_whole, good_positive, good_integer)
        self.assertEqual(five_scalars_integer_float_whole_positive, 'First, second, third, fourth, and fifth arguments are all integers or floats')
    
    def test_five_scalars_integer_float_whole_string_raises(self):
        with self.assertRaises(Exception) as context:
            five_scalars(good_integer, good_float, good_whole, good_positive, bad_scalar)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Fifth argument must be an integer or a float')

first_vector = [2, 3, 5]
second_vector = [7, 11, 13]
longer_vector = [1, 2, 3, 4]
none_vector = [None]
vector_strings = ['1 + 2k', '2 + 2k']
good_multitype = ['positive', 1, 'negative']
bad_multitype = ['positive', 'negative', 1]
bad_vector_string = 'vector'
bad_vector_buried_string = [1, 'two', 3]
bad_vector_final_string = [1, 2, 3, 4, 5, 'six']
bad_vector_nested =[[2], 3, 5]
good_long_vector = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
better_long_vector = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]

class TestConfirmVector(unittest.TestCase):
    def test_confirm_vector_scalars(self):
        confirm_vector_scalars = confirm_vector(first_vector)
        self.assertEqual(confirm_vector_scalars, 'Argument is a 1-dimensional list')
    
    def test_confirm_vector_multitype(self):
        confirm_vector_multitype = confirm_vector(good_multitype)
        self.assertEqual(confirm_vector_multitype, 'Argument is a 1-dimensional list')
    
    def test_confirm_vector_string_raises(self):
        with self.assertRaises(Exception) as context:
            confirm_vector(bad_vector_string)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Argument must be a 1-dimensional list')

    def test_confirm_vector_nested_raises(self):
        with self.assertRaises(Exception) as context:
            confirm_vector(bad_vector_nested)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Argument must be a 1-dimensional list')

class TestVectorScalars(unittest.TestCase):
    def test_vector_scalars_3long(self):
        vector_scalars_3long = vector_of_scalars(first_vector)
        self.assertEqual(vector_scalars_3long, 'Argument is a 1-dimensional list containing elements that are integers or floats')
    
    def test_vector_scalars_4long(self):
        vector_scalars_4long = vector_of_scalars(longer_vector)
        self.assertEqual(vector_scalars_4long, 'Argument is a 1-dimensional list containing elements that are integers or floats')
    
    def test_vector_scalars_none_raises(self):
        with self.assertRaises(Exception) as context:
            vector_of_scalars(none_vector)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Elements of argument must be integers or floats')
    
    def test_vector_scalars_multitype_raises(self):
        with self.assertRaises(Exception) as context:
            vector_of_scalars(good_multitype)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Elements of argument must be integers or floats')
    
    def test_vector_scalars_buried_string_raises(self):
        with self.assertRaises(Exception) as context:
            vector_of_scalars(bad_vector_buried_string)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Elements of argument must be integers or floats')
    
    def test_vector_scalars_final_string_raises(self):
        with self.assertRaises(Exception) as context:
            vector_of_scalars(bad_vector_final_string)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Elements of argument must be integers or floats')

class TestVectorStrings(unittest.TestCase):
    def test_vector_strings_normal(self):
        vector_strings_normal = vector_of_strings(vector_strings)
        self.assertEqual(vector_strings_normal, 'Argument is a 1-dimensional list containing elements that are strings or None')
    
    def test_vector_strings_none(self):
        vector_strings_none = vector_of_strings(none_vector)
        self.assertEqual(vector_strings_none, 'Argument is a 1-dimensional list containing elements that are strings or None')
    
    def test_vector_strings_nested_raises(self):
        with self.assertRaises(Exception) as context:
            allow_none_vector(bad_vector_nested)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Elements of argument must be integers, floats, strings, or None')

class TestLength(unittest.TestCase):
    def test_length_4_4_integers(self):
        length_4_4_integers = length(longer_vector, 4)
        self.assertEqual(length_4_4_integers, 'Argument contains exactly 4 elements')
    
    def test_length_3_3_integers(self):
        length_3_3_integers = length(first_vector, 3)
        self.assertEqual(length_3_3_integers, 'Argument contains exactly 3 elements')
    
    def test_length_3_3_mix(self):
        length_3_3_mix = length(good_multitype, 3)
        self.assertEqual(length_3_3_mix, 'Argument contains exactly 3 elements')
    
    def test_length_2_3_raises(self):
        with self.assertRaises(Exception) as context:
            length(first_vector, 2)
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'Argument must contain exactly 2 elements')
    
    def test_length_5_4_raises(self):
        with self.assertRaises(Exception) as context:
            length(longer_vector, 5)
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'Argument must contain exactly 5 elements')

class TestCompareVectors(unittest.TestCase):
    def test_compare_vectors_3_3(self):
        compare_vectors_3_3 = compare_vectors(first_vector, second_vector)
        self.assertEqual(compare_vectors_3_3, 'Both arguments contain the same number of elements')
    
    def test_compare_vectors_3_4_raises(self):
        with self.assertRaises(Exception) as context:
            compare_vectors(first_vector, longer_vector)
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'Both arguments must contain the same number of elements')

class TestAllowNoneVector(unittest.TestCase):
    def test_allow_none_vector_scalars(self):
        allow_none_vector_scalars = allow_none_vector(first_vector)
        self.assertEqual(allow_none_vector_scalars, 'Elements of argument are integers, floats, strings, or None')
    
    def test_allow_none_vector_strings(self):
        allow_none_vector_strings = allow_none_vector(vector_strings)
        self.assertEqual(allow_none_vector_strings, 'Elements of argument are integers, floats, strings, or None')
    
    def test_allow_none_vector_none(self):
        allow_none_vector_none = allow_none_vector(none_vector)
        self.assertEqual(allow_none_vector_none, 'Elements of argument are integers, floats, strings, or None')
    
    def test_allow_none_vector_scalars_raises(self):
        with self.assertRaises(Exception) as context:
            vector_of_strings(first_vector)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Elements of argument must be strings or None')

class TestLongVector(unittest.TestCase):
    def test_long_vector_10(self):
        long_vector_10 = long_vector(good_long_vector)
        self.assertEqual(long_vector_10, 'First argument contains at least 10 elements')
    
    def test_long_vector_11(self):
        long_vector_11 = long_vector(better_long_vector)
        self.assertEqual(long_vector_11, 'First argument contains at least 10 elements')
    
    def test_long_vector_3_raises(self):
        with self.assertRaises(Exception) as context:
            long_vector(first_vector)
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'First argument must contain at least 10 elements')

first_matrix = [[1, 2, 3], [4, 5, 6]]
second_matrix = [[7, 8, 9], [10, 11, 12]]
small_square = [[1, 2], [3, 4]]
large_square = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
matrix_points = [[1, 2], [3, 2], [5, 2], ['1 + 2k', 2]]
bad_matrix_string = 'matrix'
bad_matrix_buried_string = [['one', 2, 3], [4, 5, 6]]
bad_matrix_last_string = [[1, 2, 3], [4, 5, 6], [7, 8, 'nine']]
bad_matrix_vector = [1, 2, 3]
bad_matrix_buried_not_vector = [[1, 2, 3], [4, 5, 6], 7, [8, 9]]
bad_matrix_last_not_vector = [[1, 2, 3], [4, 5, 6], [7, 8], 9]
bad_matrix_nested = [[[1], 2, 3], [4, 5, 6]]

class TestConfirmMatrix(unittest.TestCase):
    def test_confirm_matrix_scalars(self):
        confirm_matrix_scalars = confirm_matrix(first_matrix)
        self.assertEqual(confirm_matrix_scalars, 'Argument is a 2-dimensional list')
    
    def test_confirm_matrix_string(self):
        confirm_matrix_string = confirm_matrix(bad_matrix_buried_string)
        self.assertEqual(confirm_matrix_string, 'Argument is a 2-dimensional list')
    
    def test_confirm_matrix_string_raises(self):
        with self.assertRaises(Exception) as context:
            confirm_matrix(bad_matrix_string)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Argument must be a 2-dimensional list')
    
    def test_confirm_matrix_vector_raises(self):
        with self.assertRaises(Exception) as context:
            confirm_matrix(bad_matrix_vector)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Argument must be a 2-dimensional list')

    def test_confirm_matrix_nested_raises(self):
        with self.assertRaises(Exception) as context:
            confirm_matrix(bad_matrix_nested)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Argument must be a 2-dimensional list')

class TestMatrixScalars(unittest.TestCase):
    def test_matrix_scalars_2x3(self):
        matrix_scalars_2x3 = matrix_of_scalars(first_matrix)
        self.assertEqual(matrix_scalars_2x3, 'Argument is a 2-dimensional list containing nested elements that are integers or floats')
    
    def test_matrix_scalars_2x2(self):
        matrix_scalars_2x2 = matrix_of_scalars(small_square)
        self.assertEqual(matrix_scalars_2x2, 'Argument is a 2-dimensional list containing nested elements that are integers or floats')

    def test_matrix_scalars_buried_not_vector_raises(self):
        with self.assertRaises(Exception) as context:
            matrix_of_scalars(bad_matrix_buried_not_vector)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Elements within argument must be lists')
    
    def test_matrix_scalars_last_not_vector_raises(self):
        with self.assertRaises(Exception) as context:
            matrix_of_scalars(bad_matrix_last_not_vector)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Elements within argument must be lists')
    
    def test_matrix_scalars_buried_string_raises(self):
        with self.assertRaises(Exception) as context:
            matrix_of_scalars(bad_matrix_buried_string)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Elements within lists within argument must be integers or floats')
    
    def test_matrix_scalars_last_string_raises(self):
        with self.assertRaises(Exception) as context:
            matrix_of_scalars(bad_matrix_last_string)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Elements within lists within argument must be integers or floats')

class TestSquareMatrix(unittest.TestCase):
    def test_square_matrix_2x2(self):
        square_matrix_2x2 = square_matrix(small_square)
        self.assertEqual(square_matrix_2x2, 'First argument contains the same amount of lists as the amount of elements contained within its first list')
    
    def test_square_matrix_3x3(self):
        square_matrix_3x3 = square_matrix(large_square)
        self.assertEqual(square_matrix_3x3, 'First argument contains the same amount of lists as the amount of elements contained within its first list')
    
    def test_square_matrix_2x3_raises(self):
        with self.assertRaises(Exception) as context:
            square_matrix(first_matrix)
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'First argument must contain the same amount of lists as the amount of elements contained within its first list')

class TestCompareRows(unittest.TestCase):
    def test_compare_rows_2x2_2x3(self):
        compare_rows_2x2_2x3 = compare_rows(small_square, first_matrix)
        self.assertEqual(compare_rows_2x2_2x3, 'First argument and second argument contain the same amount of lists')
    
    def test_compare_rows_3x3_2x3_raises(self):
        with self.assertRaises(Exception) as context:
            compare_rows(large_square, first_matrix)
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'First argument and second argument must contain the same amount of lists')

class TestCompareColumns(unittest.TestCase):
    def test_compare_columns_3x3_2x3(self):
        compare_columns_3x3_2x3 = compare_columns(large_square, first_matrix)
        self.assertEqual(compare_columns_3x3_2x3, 'First list within first argument and first list within second argument contain the same amount of elements')
    
    def test_compare_columns_2x2_2x3_raises(self):
        with self.assertRaises(Exception) as context:
            compare_columns(small_square, first_matrix)
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'First list within first argument and first list within second argument must contain the same amount of elements')

class TestCompareMatrices(unittest.TestCase):
    def test_compare_matrices_2x3_2x3(self):
        compare_matrices_2x3_2x3 = compare_matrices(first_matrix, second_matrix)
        self.assertEqual(compare_matrices_2x3_2x3, 'First argument and second argument contain the same amount of lists; first list within first argument and first list within second argument contain the same amount of elements')
    
    def test_compare_matrices_3x3_2x3_raises(self):
        with self.assertRaises(Exception) as context:
            compare_matrices(large_square, first_matrix)
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'First argument and second argument must contain the same amount of lists')

    def test_compare_matrices_2x2_2x3_raises(self):
        with self.assertRaises(Exception) as context:
            compare_matrices(small_square, first_matrix)
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'First list within first argument and first list within second argument must contain the same amount of elements')

class TestColumnsRows(unittest.TestCase):
    def test_columns_rows_2x2_2x3(self):
        columns_rows_2x2_2x3 = columns_rows(small_square, first_matrix)
        self.assertEqual(columns_rows_2x2_2x3, 'First list within first argument contains the same amount of elements as the amount of lists contained within second argument')
    
    def test_columns_rows_2x2_3x3_raises(self):
        with self.assertRaises(Exception) as context:
            columns_rows(small_square, large_square)
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'First list within first argument must contain the same amount of elements as the amount of lists contained within second argument')

class TestAllowNoneMatrix(unittest.TestCase):
    def test_none_matrix_none(self):
        none_matrix_none = allow_none_matrix(none_vector)
        self.assertEqual(none_matrix_none, 'Argument is either a 1-dimensional list that contains None or a 2-dimensional list containing lists in which the first nested list contains integers or floats and the last nested list begins with a string')
    
    def test_none_matrix_lists(self):
        none_matrix_lists = allow_none_matrix(matrix_points)
        self.assertEqual(none_matrix_lists, 'Argument is either a 1-dimensional list that contains None or a 2-dimensional list containing lists in which the first nested list contains integers or floats and the last nested list begins with a string')

    def test_none_matrix_no_string_raises(self):
        with self.assertRaises(Exception) as context:
            allow_none_matrix(first_matrix)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'If the first element of argument is not None, then the last element must be a list with a first element that is a string')

class TestAllowVectorMatrix(unittest.TestCase):
    def test_allow_vector_matrix_matrix(self):
        allow_vector_matrix_matrix = allow_vector_matrix(first_matrix)
        self.assertEqual(allow_vector_matrix_matrix, 'Argument is a list containing lists, integers, floats, strings, or None')
    
    def test_allow_vector_matrix_vector(self):
        allow_vector_matrix_vector = allow_vector_matrix(good_long_vector)
        self.assertEqual(allow_vector_matrix_vector, 'Argument is a list containing lists, integers, floats, strings, or None')
    
    def test_allow_vector_matrix_none(self):
        allow_vector_matrix_none = allow_vector_matrix(none_vector)
        self.assertEqual(allow_vector_matrix_none, 'Argument is a list containing lists, integers, floats, strings, or None')

    def test_allow_vector_matrix_nested_raises(self):
        with self.assertRaises(Exception) as context:
            allow_vector_matrix(bad_matrix_nested)
        self.assertEqual(type(context.exception), TypeError)
        self.assertEqual(str(context.exception), 'Argument cannot be more than a 2-dimensional list')

class TestLevel(unittest.TestCase):
    def test_level_2x3_2(self):
        level_2x2_2 = level(first_matrix, 2)
        self.assertEqual(level_2x2_2, 'Last argument is less than or equal to the length of the nested lists within the first argument')
    
    def test_level_2x3_4_raises(self):
        with self.assertRaises(Exception) as context:
            level(first_matrix, 4)
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), 'Last argument must be less than or equal to the length of the nested lists within the first argument')

good_equation = 'hyperbolic'
bad_equation = 'rational'

class TestSelectEquations(unittest.TestCase):
    def test_select_equations_included(self):
        select_equations_multiple = select_equations(good_equation)
        self.assertEqual(select_equations_multiple, "First argument is either 'linear', 'quadratic', 'cubic', 'hyperbolic', 'exponential', 'logarithmic', 'logistic', or 'sinusoidal'")
    
    def test_select_equations_excluded_raises(self):
        with self.assertRaises(Exception) as context:
            select_equations(bad_equation)
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), "First argument must be either 'linear', 'quadratic', 'cubic', 'hyperbolic', 'exponential', 'logarithmic', 'logistic', or 'sinusoidal'")

good_points = 'intercepts'
bad_points = 'criticals'

class TestSelectPoints(unittest.TestCase):
    def test_select_points_included(self):
        select_points_multiple = select_points(good_points)
        self.assertEqual(select_points_multiple, "Argument is either 'point', 'intercepts', 'maxima', 'minima', or 'inflections'")
    
    def test_select_points_excluded_raises(self):
        with self.assertRaises(Exception) as context:
            select_points(bad_points)
        self.assertEqual(type(context.exception), ValueError)
        self.assertEqual(str(context.exception), "Argument must be either 'point', 'intercepts', 'maxima', 'minima', or 'inflections'")

if __name__ == '__main__':
    unittest.main()

# ---------- Ran 111 tests in 0.007s ---------- OK ---------- #


import unittest

from library.statistics.rounding import rounded_value, rounded_list
from library.statistics.summation import sum_value
from library.statistics.sort import sorted_list, sorted_dimension, sorted_strings
from library.statistics.halve import half, half_dimension
from library.statistics.minimum import minimum_value
from library.statistics.maximum import maximum_value
from library.statistics.quartiles import quartile_value
from library.statistics.median import median_value
from library.statistics.mean import mean_value
from library.statistics.summary import five_number_summary
from library.statistics.ranges import range_value, shift_into_range
from library.statistics.deviations import multiple_deviations
from library.statistics.residuals import multiple_residuals
from library.statistics.correlation import correlation_coefficient

even_set = [8, 2, 5, 9, 1, 3, 22, 11, 9, 13]
odd_set = [7, 4, 6, 8, 2, 5, 25, 14, 8]
compare_set = [5, 5, 5, 10, 1, 7, 22, 13, 8]
counting_set = [1, 2, 3, 4, 5]
identical_set = [1, 1, 1, 1, 1]
approximate_set = [1.0001, 1.0001, 1.0001, 1.0001, 1.0001]
extreme_set = [1, 13, 59, 134, 2892]
dimension_set = [[2, 7], [1, 9], [5, 2], [3, 4], [4, 3], [8, 1], [2, 3], [7, 7], [1, 1], [5, 3]]
strings_ordered = ['2 + 3k', '4 + 3k']
strings_unordered = ['4 + 3k', '2 + 3k']

normal_decimal = 6.817239833721
extreme_decimal = 0.000000000000000000005782016894

low_precision = 2
high_precision = 8

class TestRounding(unittest.TestCase):
    def test_round_missing(self):
        round_missing = rounded_value(normal_decimal)
        self.assertEqual(round_missing, 6.8172)
    
    def test_round_normal_low(self):
        round_normal_low = rounded_value(normal_decimal, low_precision)
        self.assertEqual(round_normal_low, 6.82)

    def test_round_normal_high(self):
        round_normal_high = rounded_value(normal_decimal, high_precision)
        self.assertEqual(round_normal_high, 6.81723983)
    
    def test_round_extreme_low(self):
        round_extreme_low = rounded_value(extreme_decimal, low_precision)
        self.assertEqual(round_extreme_low, 0.01)
    
    def test_round_extreme_high(self):
        round_extreme_high = rounded_value(extreme_decimal, high_precision) 
        self.assertEqual(round_extreme_high, 1e-08)
    
    def test_rounded_list(self):
        rounded_list_main = rounded_list([3.12718492, 2.17729, 54.21, 8.9999222222, 3.9274826, 115.28191], 6)
        self.assertEqual(rounded_list_main, [3.127185, 2.17729, 54.21, 8.999922, 3.927483, 115.28191])

class TestSummation(unittest.TestCase):
    def test_sum_even(self):
        sum_even = sum_value(even_set)
        self.assertEqual(sum_even, 83)

    def test_sum_odd(self):
        sum_odd = sum_value(odd_set)
        self.assertEqual(sum_odd, 79)

class TestSort(unittest.TestCase):
    def test_sort_even(self):
        sort_even = sorted_list(even_set)
        self.assertEqual(sort_even, [1, 2, 3, 5, 8, 9, 9, 11, 13, 22])

    def test_sort_odd(self):
        sort_odd = sorted_list(odd_set)
        self.assertEqual(sort_odd, [2, 4, 5, 6, 7, 8, 8, 14, 25])
    
    def test_sort_dimension(self):
        dimension_sort = sorted_dimension(dimension_set, 1)
        self.assertEqual(dimension_sort, [[1, 9], [1, 1], [2, 7], [2, 3], [3, 4], [4, 3], [5, 2], [5, 3], [7, 7], [8, 1]])
    
    def test_sort_strings_ordered(self):
        sort_strings_ordered = sorted_strings(strings_ordered)
        self.assertEqual(sort_strings_ordered, strings_ordered)
    
    def test_sort_strings_unordered(self):
        sort_strings_unordered = sorted_strings(strings_unordered)
        self.assertEqual(sort_strings_unordered, strings_ordered)

class TestHalve(unittest.TestCase):
    def test_halve_even(self):
        halve_even = half(even_set)
        self.assertEqual(halve_even, {'upper': [9, 9, 11, 13, 22], 'lower': [1, 2, 3, 5, 8]})

    def test_halve_odd(self):
        halve_odd = half(odd_set)
        self.assertEqual(halve_odd, {'upper': [8, 8, 14, 25], 'lower': [2, 4, 5, 6]})
    
    def test_halve_dimension(self):
        dimension_halve = half_dimension(dimension_set, 1)
        self.assertEqual(dimension_halve, {'upper': [[4, 3], [5, 2], [5, 3], [7, 7], [8, 1]], 'lower': [[1, 9], [1, 1], [2, 7], [2, 3], [3, 4]]})

class TestMinimum(unittest.TestCase):
    def test_min_even(self):
        min_even = minimum_value(even_set)
        self.assertEqual(min_even, 1)

    def test_min_odd(self):
        min_odd = minimum_value(odd_set)
        self.assertEqual(min_odd, 2)

class TestMaximum(unittest.TestCase):
    def test_max_even(self):
        max_even = maximum_value(even_set)
        self.assertEqual(max_even, 22)

    def test_max_odd(self):
        max_odd = maximum_value(odd_set)
        self.assertEqual(max_odd, 25)

class TestQuartiles(unittest.TestCase):
    def test_q1_even(self):
        q1_even = quartile_value(even_set, 1)
        self.assertEqual(q1_even, 3)
    
    def test_q1_odd(self):
        q1_odd = quartile_value(odd_set, 1)
        self.assertEqual(q1_odd, 4.5)
    
    def test_q3_even(self):
        q3_even = quartile_value(even_set, 3)
        self.assertEqual(q3_even, 11)
    
    def test_q3_odd(self):
        q3_odd = quartile_value(odd_set, 3)
        self.assertEqual(q3_odd, 11)

class TestMedian(unittest.TestCase):
    def test_median_even(self):
        median_even = median_value(even_set)
        self.assertEqual(median_even, 8.5)
    
    def test_median_odd(self):
        median_odd = median_value(odd_set)
        self.assertEqual(median_odd, 7)

class TestMean(unittest.TestCase):
    def test_mean_even(self):
        mean_even = mean_value(even_set)
        self.assertEqual(mean_even, 8.3)
    
    def test_mean_odd(self):
        mean_odd = mean_value(odd_set)
        self.assertEqual(mean_odd, 8.777777777777779)

class TestFiveNumberSummary(unittest.TestCase):
    def test_five_even(self):
        five_even = five_number_summary(even_set)
        self.assertEqual(five_even, {'minimum': 1, 'q1': 3, 'median': 8.5, 'q3': 11, 'maximum': 22})
    
    def test_five_odd(self):
        five_odd = five_number_summary(odd_set)
        self.assertEqual(five_odd, {'minimum': 2, 'q1': 4.5, 'median': 7, 'q3': 11.0, 'maximum': 25})

class TestRange(unittest.TestCase):
    def test_range_even(self):
        range_even = range_value(even_set)
        self.assertEqual(range_even, 21)
    
    def test_range_odd(self):
        range_odd = range_value(odd_set)
        self.assertEqual(range_odd, 23)

class TestShiftIntoRange(unittest.TestCase):
    def test_shift_range_within_positive(self):
        shift_range_within_positive = shift_into_range(11, 2, 10, 20)
        self.assertEqual(shift_range_within_positive, 11)
    
    def test_shift_range_at_min_positive(self):
        shift_range_at_min_positive = shift_into_range(10, 2, 10, 20)
        self.assertEqual(shift_range_at_min_positive, 10)
    
    def test_shift_range_at_max_positive(self):
        shift_range_at_max_positive = shift_into_range(20, 2, 10, 20)
        self.assertEqual(shift_range_at_max_positive, 20)
    
    def test_shift_range_below_positive(self):
        shift_range_below_positive = shift_into_range(1, 2, 10, 20)
        self.assertEqual(shift_range_below_positive, 11)
    
    def test_shift_range_above_positive(self):
        shift_range_above_positive = shift_into_range(31, 2, 10, 20)
        self.assertEqual(shift_range_above_positive, 19)
    
    def test_shift_range_unfit_above_positive(self):
        shift_range_unfit_above_positive = shift_into_range(31, 30, 10, 20)
        self.assertEqual(shift_range_unfit_above_positive, 31)
    
    def test_shift_range_unfit_below_positive(self):
        shift_range_unfit_below_positive = shift_into_range(1, 30, 10, 20)
        self.assertEqual(shift_range_unfit_below_positive, 31)
    
    def test_shift_range_within_negative(self):
        shift_range_within_negative = shift_into_range(11, -2, 10, 20)
        self.assertEqual(shift_range_within_negative, 11)
    
    def test_shift_range_at_min_negative(self):
        shift_range_at_min_negative = shift_into_range(10, -2, 10, 20)
        self.assertEqual(shift_range_at_min_negative, 10)
    
    def test_shift_range_at_max_negative(self):
        shift_range_at_max_negative = shift_into_range(20, -2, 10, 20)
        self.assertEqual(shift_range_at_max_negative, 20)
    
    def test_shift_range_below_negative(self):
        shift_range_below_negative = shift_into_range(1, -2, 10, 20)
        self.assertEqual(shift_range_below_negative, 11)
    
    def test_shift_range_above_negative(self):
        shift_range_above_negative = shift_into_range(31, -2, 10, 20)
        self.assertEqual(shift_range_above_negative, 19)
    
    def test_shift_range_unfit_above_negative(self):
        shift_range_unfit_above_negative = shift_into_range(31, -30, 10, 20)
        self.assertEqual(shift_range_unfit_above_negative, 31)
    
    def test_shift_range_unfit_below_negative(self):
        shift_range_unfit_below_negative = shift_into_range(1, -30, 10, 20)
        self.assertEqual(shift_range_unfit_below_negative, 31)

class TestDeviations(unittest.TestCase):
    def test_deviations_even(self):
        deviations_even = multiple_deviations(even_set)
        self.assertEqual(deviations_even, [-0.3000000000000007, -6.300000000000001, -3.3000000000000007, 0.6999999999999993, -7.300000000000001, -5.300000000000001, 13.7, 2.6999999999999993, 0.6999999999999993, 4.699999999999999])
    
    def test_deviations_odd(self):
        deviations_odd = multiple_deviations(odd_set)
        self.assertEqual(deviations_odd, [-1.7777777777777786, -4.777777777777779, -2.7777777777777786, -0.7777777777777786, -6.777777777777779, -3.7777777777777786, 16.22222222222222, 5.222222222222221, -0.7777777777777786])
    
    def test_deviations_extreme(self):
        deviations_extreme = multiple_deviations(extreme_set)
        self.assertEqual(deviations_extreme, [-618.8, -606.8, -560.8, -485.79999999999995, 2272.2])

class TestResiduals(unittest.TestCase):
    def test_residuals_compare(self):
        residuals_compare = multiple_residuals(odd_set, compare_set)
        self.assertEqual(residuals_compare, [2, -1, 1, -2, 1, -2, 3, 1, 0])
    
    def test_residuals_extreme(self):
        residuals_extreme = multiple_residuals(extreme_set, counting_set)
        self.assertEqual(residuals_extreme, [0.0, 11.0, 56.0, 130.0, 2887.0])

class TestCorrelation(unittest.TestCase):    
    def test_correlation_compare(self):
        correlation_compare = correlation_coefficient(odd_set, compare_set)
        self.assertEqual(correlation_compare, 0.967)
    
    def test_correlation_deviation_zero(self):
        correlation_deviation_zero = correlation_coefficient(identical_set, approximate_set)
        self.assertEqual(correlation_deviation_zero, 0.9997)
    
    def test_correlation_ratio_over(self):
        correlation_ratio_over = correlation_coefficient(extreme_set, identical_set)
        self.assertEqual(correlation_ratio_over, 0.0)

if __name__ == '__main__':
    unittest.main()

# ---------- Ran 54 tests in 0.005s ---------- OK ---------- #